from ....c99 import (
    Clearerr,
    Fclose,
    Feof,
    Ferror,
    Fflush,
    Fgetc,
    Fgetpos,
    Fgets,
    Fopen,
    Fprintf,
    Fputc,
    Fputs,
    Fread,
    Freopen,
    Fscanf,
    Fseek,
    Fsetpos,
    Ftell,
    Fwrite,
    Getc,
    Getchar,
    Gets,
    Printf,
    Putc,
    Putchar,
    Puts,
    Remove,
    Rename,
    Rewind,
    Scanf,
    Snprintf,
    Sprintf,
    Sscanf,
    Tmpfile,
    Tmpnam,
    Ungetc,
    Vfprintf,
    Vfscanf,
    Vprintf,
    Vscanf,
    Vsnprintf,
    Vsprintf,
    Vsscanf,
)
from ..systemv import RiscV64SysVModel


class RiscV64SysVFclose(Fclose, RiscV64SysVModel):
    pass


class RiscV64SysVFeof(Feof, RiscV64SysVModel):
    pass


class RiscV64SysVFerror(Ferror, RiscV64SysVModel):
    pass


class RiscV64SysVFgetc(Fgetc, RiscV64SysVModel):
    pass


class RiscV64SysVFgets(Fgets, RiscV64SysVModel):
    pass


class RiscV64SysVFopen(Fopen, RiscV64SysVModel):
    pass


class RiscV64SysVFprintf(Fprintf, RiscV64SysVModel):
    pass


class RiscV64SysVFputc(Fputc, RiscV64SysVModel):
    pass


class RiscV64SysVFputs(Fputs, RiscV64SysVModel):
    pass


class RiscV64SysVFread(Fread, RiscV64SysVModel):
    pass


class RiscV64SysVFscanf(Fscanf, RiscV64SysVModel):
    pass


class RiscV64SysVFseek(Fseek, RiscV64SysVModel):
    pass


class RiscV64SysVFtell(Ftell, RiscV64SysVModel):
    pass


class RiscV64SysVFwrite(Fwrite, RiscV64SysVModel):
    pass


class RiscV64SysVGetc(Getc, RiscV64SysVModel):
    pass


class RiscV64SysVGetchar(Getchar, RiscV64SysVModel):
    pass


class RiscV64SysVGets(Gets, RiscV64SysVModel):
    pass


class RiscV64SysVPrintf(Printf, RiscV64SysVModel):
    pass


class RiscV64SysVPutc(Putc, RiscV64SysVModel):
    pass


class RiscV64SysVPutchar(Putchar, RiscV64SysVModel):
    pass


class RiscV64SysVPuts(Puts, RiscV64SysVModel):
    pass


class RiscV64SysVRemove(Remove, RiscV64SysVModel):
    pass


class RiscV64SysVRename(Rename, RiscV64SysVModel):
    pass


class RiscV64SysVRewind(Rewind, RiscV64SysVModel):
    pass


class RiscV64SysVScanf(Scanf, RiscV64SysVModel):
    pass


class RiscV64SysVSnprintf(Snprintf, RiscV64SysVModel):
    pass


class RiscV64SysVSprintf(Sprintf, RiscV64SysVModel):
    pass


class RiscV64SysVSscanf(Sscanf, RiscV64SysVModel):
    pass


class RiscV64SysVClearerr(Clearerr, RiscV64SysVModel):
    pass


class RiscV64SysVFflush(Fflush, RiscV64SysVModel):
    pass


class RiscV64SysVFreopen(Freopen, RiscV64SysVModel):
    pass


class RiscV64SysVFgetpos(Fgetpos, RiscV64SysVModel):
    pass


class RiscV64SysVFsetpos(Fsetpos, RiscV64SysVModel):
    pass


class RiscV64SysVTmpfile(Tmpfile, RiscV64SysVModel):
    pass


class RiscV64SysVTmpnam(Tmpnam, RiscV64SysVModel):
    pass


class RiscV64SysVUngetc(Ungetc, RiscV64SysVModel):
    pass


class RiscV64SysVVfprintf(Vfprintf, RiscV64SysVModel):
    pass


class RiscV64SysVVfscanf(Vfscanf, RiscV64SysVModel):
    pass


class RiscV64SysVVprintf(Vprintf, RiscV64SysVModel):
    pass


class RiscV64SysVVscanf(Vscanf, RiscV64SysVModel):
    pass


class RiscV64SysVVsnprintf(Vsnprintf, RiscV64SysVModel):
    pass


class RiscV64SysVVsprintf(Vsprintf, RiscV64SysVModel):
    pass


class RiscV64SysVVsscanf(Vsscanf, RiscV64SysVModel):
    pass


__all__ = [
    "RiscV64SysVFclose",
    "RiscV64SysVFeof",
    "RiscV64SysVFerror",
    "RiscV64SysVFgetc",
    "RiscV64SysVFgets",
    "RiscV64SysVFopen",
    "RiscV64SysVFprintf",
    "RiscV64SysVFputc",
    "RiscV64SysVFputs",
    "RiscV64SysVFread",
    "RiscV64SysVFscanf",
    "RiscV64SysVFseek",
    "RiscV64SysVFtell",
    "RiscV64SysVFwrite",
    "RiscV64SysVGetc",
    "RiscV64SysVGetchar",
    "RiscV64SysVGets",
    "RiscV64SysVPrintf",
    "RiscV64SysVPutc",
    "RiscV64SysVPutchar",
    "RiscV64SysVPuts",
    "RiscV64SysVRemove",
    "RiscV64SysVRename",
    "RiscV64SysVRewind",
    "RiscV64SysVScanf",
    "RiscV64SysVSnprintf",
    "RiscV64SysVSprintf",
    "RiscV64SysVSscanf",
    "RiscV64SysVClearerr",
    "RiscV64SysVFflush",
    "RiscV64SysVFreopen",
    "RiscV64SysVFgetpos",
    "RiscV64SysVFsetpos",
    "RiscV64SysVTmpfile",
    "RiscV64SysVTmpnam",
    "RiscV64SysVUngetc",
    "RiscV64SysVVfprintf",
    "RiscV64SysVVfscanf",
    "RiscV64SysVVprintf",
    "RiscV64SysVVscanf",
    "RiscV64SysVVsnprintf",
    "RiscV64SysVVsprintf",
    "RiscV64SysVVsscanf",
]



import typing

from .emulator import (
    Emulator,
    FunctionHookable,
    InstructionHookable,
    InterruptHookable,
    MemoryReadHookable,
    MemoryWriteHookable,
)


def range_intersect(r1, r2):
    return range(max(r1.start, r2.start), min(r1.stop, r2.stop)) or None


def range_to_hex_str(r):
    return f"[{r.start:x}..{r.stop:x}]"


class QInstructionHookable(InstructionHookable):
    """Implementation of instruction hook data structure mgmt to be used by qemu-based emulators."""

    def __init__(self: typing.Any) -> None:
        super().__init__()
        self.instruction_hooks: typing.Dict[int, typing.Callable[[Emulator], None]] = {}
        self.all_instructions_hook: typing.Optional[
            typing.Callable[[Emulator], None]
        ] = None

    def hook_instructions(self, function: typing.Callable[[Emulator], None]) -> None:
        self.all_instructions_hook = function

    def unhook_instructions(self):
        self.all_instructions_hook = None

    def hook_instruction(
        self, address: int, function: typing.Callable[[Emulator], None]
    ) -> None:
        if address in self.instruction_hooks:
            raise ValueError(
                f"can't hook instruction at {address:x} since already hooked"
            )
        self.instruction_hooks[address] = function

    def unhook_instruction(self, address: int) -> None:
        if address not in self.instruction_hooks:
            raise ValueError(
                f"can't unhook instruction at {address:x} since its not already hooked"
            )
        self.instruction_hooks.pop(address, None)

    def is_instruction_hooked(
        self, address: int
    ) -> typing.Optional[typing.Callable[[Emulator], None]]:
        if address in self.instruction_hooks:
            return self.instruction_hooks[address]
        else:
            return None


class QFunctionHookable(FunctionHookable):
    """Implementation of function hook data structure mgmt to be used by qemu-based emulators."""

    def __init__(self: typing.Any) -> None:
        super().__init__()
        self.function_hooks: typing.Dict[int, typing.Callable[[Emulator], None]] = {}

    def hook_function(
        self, address: int, function: typing.Callable[[Emulator], None]
    ) -> None:
        if address in self.function_hooks:
            raise ValueError(f"can't hook function at {address:x} since already hooked")
        self.function_hooks[address] = function

    def unhook_function(self, address: int) -> None:
        if address not in self.function_hooks:
            raise ValueError(
                f"can't unhook function at {address:x} since its not already hooked"
            )
        self.function_hooks.pop(address, None)

    def is_function_hooked(
        self, address: int
    ) -> typing.Optional[typing.Callable[[Emulator], None]]:
        if address in self.function_hooks:
            return self.function_hooks[address]
        else:
            return None


class QMemoryReadHookable(MemoryReadHookable):
    """Implementation of memory read hook data structure mgmt to be used by qemu-based emulators."""

    def __init__(self: typing.Any) -> None:
        super().__init__()
        self.memory_read_hooks: typing.Dict[
            range, typing.Callable[[Emulator, int, int, bytes], typing.Optional[bytes]]
        ] = {}
        self.all_reads_hook: typing.Optional[
            typing.Callable[[Emulator, int, int, bytes], typing.Optional[bytes]]
        ] = None

    def hook_memory_reads(
        self,
        function: typing.Callable[[Emulator, int, int, bytes], typing.Optional[bytes]],
    ) -> None:
        self.all_reads_hook = function

    def unhook_memory_reads(self):
        self.all_reads_hook = None

    def hook_memory_read(
        self,
        start: int,
        end: int,
        function: typing.Callable[[Emulator, int, int, bytes], typing.Optional[bytes]],
    ) -> None:
        new_range = range(start, end)
        for r in self.memory_read_hooks:
            if range_intersect(r, new_range):
                raise ValueError(
                    f"can't hook memory read range {range_to_hex_str(new_range)} since that overlaps already hooked range {range_to_hex_str(r)}"
                )
        self.memory_read_hooks[new_range] = function

    def unhook_memory_read(self, start: int, end: int) -> None:
        unh_r = range(start, end)
        for r in self.memory_read_hooks:
            if unh_r == r:
                self.memory_read_hooks.pop(r, None)
                return
        raise ValueError(
            f"can't unhook memory read range {range_to_hex_str(unh_r)} since its not already hooked"
        )

    # def unhook_memory_read(address: int) -> None:
    #    if address not in self.memory_read_hooks:
    #        raise ValueError(
    #            f"can't unhook memory read range {range_to_hex_str(address)} since its not already hooked"
    #        )
    #    self.memory_read_hooks.pop(address, None)

    def is_memory_read_hooked(
        self, address: int, size: int
    ) -> typing.Optional[
        typing.Callable[[Emulator, int, int, bytes], typing.Optional[bytes]]
    ]:
        end = address + size
        access_rng = range(address, end)
        for rng in self.memory_read_hooks:
            if (
                address in rng
                or end - 1 in rng
                or rng.start in access_rng
                or rng.stop - 1 in access_rng
            ):
                return self.memory_read_hooks[rng]
        return None


class QMemoryWriteHookable(MemoryWriteHookable):
    """Implementation of memory write hook data structure mgmt to be used by qemu-based emulators."""

    def __init__(self: typing.Any) -> None:
        super().__init__()
        self.memory_write_hooks: typing.Dict[
            range, typing.Callable[[Emulator, int, int, bytes], None]
        ] = {}
        self.all_writes_hook: typing.Optional[
            typing.Callable[[Emulator, int, int, bytes], None]
        ] = None

    def hook_memory_writes(
        self, function: typing.Callable[[Emulator, int, int, bytes], None]
    ) -> None:
        self.all_writes_hook = function

    def unhook_memory_writes(self):
        self.all_writes_hook = None

    def hook_memory_write(
        self,
        start: int,
        end: int,
        function: typing.Callable[[Emulator, int, int, bytes], None],
    ) -> None:
        new_range = range(start, end)
        for r in self.memory_write_hooks:
            if range_intersect(r, new_range):
                raise ValueError(
                    f"can't hook memory write range {range_to_hex_str(new_range)} since that overlaps already hooked range {range_to_hex_str(r)}"
                )
        self.memory_write_hooks[new_range] = function

    def unhook_memory_write(self, start: int, end: int) -> None:
        unh_r = range(start, end)
        for r in self.memory_write_hooks:
            if unh_r == r:
                self.memory_write_hooks.pop(r, None)
                return
        raise ValueError(
            f"can't unhook memory write range {range_to_hex_str(unh_r)} since its not already hooked"
        )

    # def unhook_memory_write(self, address: int) -> None:
    #    if address not in self.memory_write_hooks:
    #        raise ValueError(
    #            f"can't unhook memory write range {range_to_hex_str(address)} since its not already hooked"
    #        )
    #    self.memory_write_hooks.pop(address, None)
    def is_memory_write_hooked(
        self, address: int, size: int
    ) -> typing.Optional[typing.Callable[[Emulator, int, int, bytes], None]]:
        end = address + size
        access_rng = range(address, end)
        for rng in self.memory_write_hooks:
            if (
                address in rng
                or end - 1 in rng
                or rng.start in access_rng
                or rng.stop - 1 in access_rng
            ):
                return self.memory_write_hooks[rng]
        return None


class QInterruptHookable(InterruptHookable):
    """Implementation of interrupt hook data structure mgmt to be used by qemu-based emulators."""

    def __init__(self: typing.Any) -> None:
        super().__init__()
        self.all_interrupts_hook: typing.Optional[
            typing.Callable[[Emulator, int], None]
        ] = None
        self.interrupt_hooks: typing.Dict[int, typing.Callable[[Emulator], None]] = {}

    def hook_interrupts(self, function: typing.Callable[[Emulator, int], None]) -> None:
        self.all_interrupts_hook = function

    def unhook_interrupts(self):
        self.all_interrupts_hook = None

    def hook_interrupt(
        self, intno: int, function: typing.Callable[[Emulator], None]
    ) -> None:
        if intno in self.interrupt_hooks:
            raise ValueError(
                f"can't hook interrupt number {intno} since its already hooked"
            )
        self.interrupt_hooks[intno] = function

    def unhook_interrupt(self, intno: int) -> None:
        if intno not in self.interrupt_hooks:
            raise ValueError(
                f"asked to unhook interrupt {intno} which was not previously hooked"
            )
        self.interrupt_hooks.pop(intno, None)

    def is_interrupt_hooked(
        self, intno: int
    ) -> typing.Optional[typing.Callable[[Emulator], None]]:
        if intno in self.interrupt_hooks:
            return self.interrupt_hooks[intno]
        else:
            return None



import logging
import re

import networkx as nx

from .. import hinting
from . import analysis

logger = logging.getLogger(__name__)


class DefUseGraph(nx.MultiDiGraph):
    def add_def_use(self, def_node, use_node, def_info, use_info, color):
        # logger.info(f"def use def_node={def_node} use_node={use_node}")
        # logger.info(f"  def_info={def_info}")
        # logger.info(f"  use_info={use_info}")
        self.add_edges_from(
            [
                (
                    def_node,
                    use_node,
                    {"def_info": def_info, "use_info": use_info, "color": color},
                )
            ]
        )


class ColorizerDefUse(analysis.Analysis):
    name = "colorizer_def_use_graph"
    description = "assemble a def use graph from colorizer summary hints"
    version = "0.0.1"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.new_hints = []
        self.not_new_hints = []
        self.hinter.register(
            hinting.DynamicRegisterValueSummaryHint, self.collect_hints
        )
        self.hinter.register(hinting.DynamicMemoryValueSummaryHint, self.collect_hints)

    def collect_hints(self, hint: hinting.Hint):
        if (
            type(hint) is hinting.DynamicRegisterValueSummaryHint
            or type(hint) is hinting.DynamicMemoryValueSummaryHint
        ):
            if hint.new:
                self.new_hints.append(hint)
            else:
                self.not_new_hints.append(hint)

    def run(self, machine):
        self.du_graph = DefUseGraph()
        color2genesis = {}

        def hint_dv_info(hint):
            if type(hint) is hinting.DynamicRegisterValueSummaryHint:
                return {
                    "new": hint.new,
                    "count": hint.count,
                    "num_micro_executions": hint.num_micro_executions,
                    "type": "reg",
                    "is_read": hint.use,
                    "reg_name": hint.reg_name,
                    "color": hint.color,
                }
            elif type(hint) is hinting.DynamicMemoryValueSummaryHint:
                return {
                    "new": hint.new,
                    "count": hint.count,
                    "num_micro_executions": hint.num_micro_executions,
                    "type": "mem",
                    "is_read": hint.use,
                    "base": hint.base,
                    "index": hint.index,
                    "scale": hint.scale,
                    "offset": hint.offset,
                    "color": hint.color,
                }
            else:
                assert 1 == 0

        for hint in self.new_hints:
            # this is a new color so its either an input or a write of a computed value

            # some sanity checking
            assert "def" in hint.message
            assert hint.color not in color2genesis
            assert (hint.use and ("read" in hint.message)) or (
                (not hint.use) and ("write" in hint.message)
            )

            self.du_graph.add_node(hint.pc)

            dv_info = hint_dv_info(hint)

            if hint.use:
                # this is a read
                # hallucinate a node representing the creation of that color
                color_node = f"input color-{hint.color}"
                assert dv_info["type"] == "reg" or dv_info["type"] == "mem"

                if dv_info["type"] == "reg":
                    color_node = color_node + " " + dv_info["reg_name"]
                if dv_info["type"] == "mem":
                    color_node = color_node + " *(" + dv_info["base"]
                    if dv_info["index"] != "None":
                        color_node += f"{dv_info['scale']}*{dv_info['index']}"
                    if dv_info["offset"] > 0:
                        color_node += f"{dv_info['offset']}"
                    color_node += ")"
                color_node += "_init"
                self.du_graph.add_node(color_node)
                # record mapping from this color to its creation info
                color2genesis[hint.color] = (color_node, dv_info)
                # and an edge between that color node and this instruction
                self.du_graph.add_def_use(
                    color_node, hint.pc, color_node, dv_info, hint.color
                )
            else:
                # this is a write of a computed value
                # record mapping from this color to its creation info
                color2genesis[hint.color] = (hint.pc, dv_info)

        for hint in self.not_new_hints:
            # not a new color.  so its a flow
            self.du_graph.add_node(hint.pc)
            # can't be a def
            assert "def" not in hint.message
            # we should never see !new && !use since that is just a value copy
            # assert hint.use
            if not hint.use:
                pass
            else:
                dv_info = hint_dv_info(hint)
                (def_node, def_info) = color2genesis[hint.color]
                self.du_graph.add_def_use(
                    def_node, hint.pc, def_info, dv_info, hint.color
                )

        # hint out the def-use graph
        self.hinter.send(
            hinting.DefUseGraphHint(
                graph=nx.node_link_data(self.du_graph, edges="links"),
                message="concrete-summary-def-use-graph",
            )
        )

        with open("colorizer_def_use.dot", "w") as dot:

            def writeit(foo):
                dot.write(foo + "\n")

            writeit("digraph{")
            writeit(" rankdir=LR")

            node2nodeid = {}
            for node in self.du_graph.nodes:
                # 4461 is pc
                # {"id": 4461},
                # An input color rsp is register
                # {"id": "input color-1 rsp_init"},
                node_id = f"node_{len(node2nodeid)}"
                if type(node) is int:
                    # nodeinfo = ("instruction", f"0x{node}")
                    writeit(f'  {node_id} [label="0x{node:x}"]')
                else:
                    assert type(node) is str
                    foo = re.search("input color-([0-9]+) (.*)_init", node)
                    assert foo is not None
                    (cns, reg) = foo.groups()
                    cn = int(cns)
                    # nodeinfo = ("input", (cn, reg))
                    writeit(f'  {node_id} [color="blue", label="input({reg})"]')
                node2nodeid[node] = node_id

            di = nx.get_edge_attributes(self.du_graph, "def_info")
            ui = nx.get_edge_attributes(self.du_graph, "use_info")
            for e in self.du_graph.edges:
                if type(di[e]) is str:
                    foo = re.search("color-([0-9]+) (.*)_init", di[e])
                    (cns, reg) = foo.groups()
                    cn = int(cns)
                    assert cn == ui[e]["color"]
                else:
                    assert di[e]["color"] == ui[e]["color"]
                (src, dst, k) = e
                cn = ui[e]["color"]
                tl = ""

                def i2s(inf):
                    lab = ""
                    if "type" in inf:
                        if inf["type"] == "reg":
                            lab = inf["reg_name"]
                        if inf["type"] == "mem":
                            lab = "["
                            if inf["base"] != "None":
                                lab += inf["base"]
                            if inf["index"] != "None":
                                lab += f'+{inf["scale"]}*{inf["index"]}'
                            if inf["offset"] != 0:
                                lab += f'+{inf["offset"]:x}'
                            lab += "]"
                    return lab

                hl = i2s(ui[e])
                tl = i2s(di[e])

                writeit(
                    f'  {node2nodeid[src]} -> {node2nodeid[dst]} [label="{cn}",headlabel="{hl}",taillabel="{tl}"]'
                )

            writeit("}\n")

    def get_graph(self):
        return self.du_graph



import struct

from ..... import emulators, platforms
from ...cstd import ArgumentType, CStdModel


class ArmHFSysVModel(CStdModel):
    """Base class for C models using the ARM32 GNU EABI

    This is a specific ARM System V ABI.
    It generally applies to ARMv5t or ARMv6.
    """

    platform = platforms.Platform(
        platforms.Architecture.ARM_V7A, platforms.Byteorder.LITTLE
    )
    abi = platforms.ABI.SYSTEMV

    _int_sign_mask = 0x80000000
    _int_inv_mask = 0xFFFFFFFF
    _long_sign_mask = 0x80000000
    _long_inv_mask = 0xFFFFFFFF
    _long_long_sign_mask = 0x8000000000000000
    _long_long_inv_mask = 0xFFFFFFFFFFFFFFFF

    _four_byte_types = {
        ArgumentType.INT,
        ArgumentType.UINT,
        ArgumentType.LONG,
        ArgumentType.ULONG,
        ArgumentType.SIZE_T,
        ArgumentType.SSIZE_T,
        ArgumentType.POINTER,
    }

    _eight_byte_types = {
        ArgumentType.LONGLONG,
        ArgumentType.ULONGLONG,
    }

    _four_byte_arg_regs = ["r0", "r1", "r2", "r3"]
    _eight_byte_arg_regs = ["r0", "r1", "r2", "r3"]

    _soft_float = False
    _variadic_soft_float = True
    _floats_are_doubles = False
    _float_arg_regs = ["s0", "s1", "s2", "s3", "s4", "s5", "s6"]
    _double_arg_regs = ["d0", "d1", "d2", "d3", "d4", "d5", "d6"]

    _init_stack_offset = 0
    _align_stack = True
    _eight_byte_reg_size = 2
    _double_reg_size = 1
    _four_byte_stack_size = 4
    _eight_byte_stack_size = 8
    _float_stack_size = 4
    _double_stack_size = 8

    def _return_4_byte(self, emulator: emulators.Emulator, val: int) -> None:
        emulator.write_register("r0", val)

    def _return_8_byte(self, emulator: emulators.Emulator, val: int) -> None:
        lo = val & self._int_inv_mask
        hi = val >> 32 & self._int_inv_mask

        emulator.write_register("r0", lo)
        emulator.write_register("r1", hi)

    def _return_float(self, emulator: emulators.Emulator, val: float) -> None:
        data = struct.pack("<f", val)
        intval = int.from_bytes(data, "little")
        emulator.write_register("s0", intval)

    def _return_double(self, emulator: emulators.Emulator, val: float) -> None:
        data = struct.pack("<d", val)
        intval = int.from_bytes(data, "little")
        emulator.write_register("d0", intval)



from ....posix.libgen import Basename, Dirname
from ..systemv import MIPS64ELSysVModel


class MIPS64ELSysVBasename(Basename, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVDirname(Dirname, MIPS64ELSysVModel):
    pass


__all__ = [
    "MIPS64ELSysVBasename",
    "MIPS64ELSysVDirname",
]



from ....c99 import (
    Clearerr,
    Fclose,
    Feof,
    Ferror,
    Fflush,
    Fgetc,
    Fgetpos,
    Fgets,
    Fopen,
    Fprintf,
    Fputc,
    Fputs,
    Fread,
    Freopen,
    Fscanf,
    Fseek,
    Fsetpos,
    Ftell,
    Fwrite,
    Getc,
    Getchar,
    Gets,
    Printf,
    Putc,
    Putchar,
    Puts,
    Remove,
    Rename,
    Rewind,
    Scanf,
    Snprintf,
    Sprintf,
    Sscanf,
    Tmpfile,
    Tmpnam,
    Ungetc,
    Vfprintf,
    Vfscanf,
    Vprintf,
    Vscanf,
    Vsnprintf,
    Vsprintf,
    Vsscanf,
)
from ..systemv import ArmELSysVModel


class ArmELSysVFclose(Fclose, ArmELSysVModel):
    pass


class ArmELSysVFeof(Feof, ArmELSysVModel):
    pass


class ArmELSysVFerror(Ferror, ArmELSysVModel):
    pass


class ArmELSysVFgetc(Fgetc, ArmELSysVModel):
    pass


class ArmELSysVFgets(Fgets, ArmELSysVModel):
    pass


class ArmELSysVFopen(Fopen, ArmELSysVModel):
    pass


class ArmELSysVFprintf(Fprintf, ArmELSysVModel):
    pass


class ArmELSysVFputc(Fputc, ArmELSysVModel):
    pass


class ArmELSysVFputs(Fputs, ArmELSysVModel):
    pass


class ArmELSysVFread(Fread, ArmELSysVModel):
    pass


class ArmELSysVFscanf(Fscanf, ArmELSysVModel):
    pass


class ArmELSysVFseek(Fseek, ArmELSysVModel):
    pass


class ArmELSysVFtell(Ftell, ArmELSysVModel):
    pass


class ArmELSysVFwrite(Fwrite, ArmELSysVModel):
    pass


class ArmELSysVGetc(Getc, ArmELSysVModel):
    pass


class ArmELSysVGetchar(Getchar, ArmELSysVModel):
    pass


class ArmELSysVGets(Gets, ArmELSysVModel):
    pass


class ArmELSysVPrintf(Printf, ArmELSysVModel):
    pass


class ArmELSysVPutc(Putc, ArmELSysVModel):
    pass


class ArmELSysVPutchar(Putchar, ArmELSysVModel):
    pass


class ArmELSysVPuts(Puts, ArmELSysVModel):
    pass


class ArmELSysVRemove(Remove, ArmELSysVModel):
    pass


class ArmELSysVRename(Rename, ArmELSysVModel):
    pass


class ArmELSysVRewind(Rewind, ArmELSysVModel):
    pass


class ArmELSysVScanf(Scanf, ArmELSysVModel):
    pass


class ArmELSysVSnprintf(Snprintf, ArmELSysVModel):
    pass


class ArmELSysVSprintf(Sprintf, ArmELSysVModel):
    pass


class ArmELSysVSscanf(Sscanf, ArmELSysVModel):
    pass


class ArmELSysVClearerr(Clearerr, ArmELSysVModel):
    pass


class ArmELSysVFflush(Fflush, ArmELSysVModel):
    pass


class ArmELSysVFreopen(Freopen, ArmELSysVModel):
    pass


class ArmELSysVFgetpos(Fgetpos, ArmELSysVModel):
    pass


class ArmELSysVFsetpos(Fsetpos, ArmELSysVModel):
    pass


class ArmELSysVTmpfile(Tmpfile, ArmELSysVModel):
    pass


class ArmELSysVTmpnam(Tmpnam, ArmELSysVModel):
    pass


class ArmELSysVUngetc(Ungetc, ArmELSysVModel):
    pass


class ArmELSysVVfprintf(Vfprintf, ArmELSysVModel):
    pass


class ArmELSysVVfscanf(Vfscanf, ArmELSysVModel):
    pass


class ArmELSysVVprintf(Vprintf, ArmELSysVModel):
    pass


class ArmELSysVVscanf(Vscanf, ArmELSysVModel):
    pass


class ArmELSysVVsnprintf(Vsnprintf, ArmELSysVModel):
    pass


class ArmELSysVVsprintf(Vsprintf, ArmELSysVModel):
    pass


class ArmELSysVVsscanf(Vsscanf, ArmELSysVModel):
    pass


__all__ = [
    "ArmELSysVFclose",
    "ArmELSysVFeof",
    "ArmELSysVFerror",
    "ArmELSysVFgetc",
    "ArmELSysVFgets",
    "ArmELSysVFopen",
    "ArmELSysVFprintf",
    "ArmELSysVFputc",
    "ArmELSysVFputs",
    "ArmELSysVFread",
    "ArmELSysVFscanf",
    "ArmELSysVFseek",
    "ArmELSysVFtell",
    "ArmELSysVFwrite",
    "ArmELSysVGetc",
    "ArmELSysVGetchar",
    "ArmELSysVGets",
    "ArmELSysVPrintf",
    "ArmELSysVPutc",
    "ArmELSysVPutchar",
    "ArmELSysVPuts",
    "ArmELSysVRemove",
    "ArmELSysVRename",
    "ArmELSysVRewind",
    "ArmELSysVScanf",
    "ArmELSysVSnprintf",
    "ArmELSysVSprintf",
    "ArmELSysVSscanf",
    "ArmELSysVClearerr",
    "ArmELSysVFflush",
    "ArmELSysVFreopen",
    "ArmELSysVFgetpos",
    "ArmELSysVFsetpos",
    "ArmELSysVTmpfile",
    "ArmELSysVTmpnam",
    "ArmELSysVUngetc",
    "ArmELSysVVfprintf",
    "ArmELSysVVfscanf",
    "ArmELSysVVprintf",
    "ArmELSysVVscanf",
    "ArmELSysVVsnprintf",
    "ArmELSysVVsprintf",
    "ArmELSysVVsscanf",
]



from ..... import platforms
from .....exceptions import ConfigurationError
from ..structs import ElfRela
from .rela import ElfRelocator

R_PPC_ABS32 = 1  # Direct 32-bit
R_PPC_GLOB_DAT = 20  # Create GOT entry
R_PPC_JUMP_SLOT = 21  # Create PLT entry
R_PPC_RELATIVE = 22  # Adjust by program base


class PowerPCElfRelocator(ElfRelocator):
    arch = platforms.Architecture.POWERPC32
    byteorder = platforms.Byteorder.BIG
    addrsz = 4

    def _compute_value(self, rela: ElfRela):
        if rela.type == R_PPC_ABS32:
            val = rela.symbol.value + rela.symbol.baseaddr + rela.addend
            return val.to_bytes(4, "big")
        elif (
            rela.type == R_PPC_GLOB_DAT
            or rela.type == R_PPC_JUMP_SLOT
            or rela.type == R_PPC_RELATIVE
        ):
            # Different semantics, all behave the same
            val = rela.symbol.value + rela.symbol.baseaddr + rela.addend
            return val.to_bytes(self.addrsz, "big")
        else:
            raise ConfigurationError(
                f"Invalid relocation type for {rela.symbol.name}: {rela.type}"
            )


R_PPC64_ADDR64 = 38  # Adjust by program base, I think...


class PowerPC64ElfRelocator(PowerPCElfRelocator):
    arch = platforms.Architecture.POWERPC64
    byteorder = platforms.Byteorder.BIG
    addrsz = 8

    def _compute_value(self, rela: ElfRela):
        if rela.type == R_PPC64_ADDR64:
            val = rela.symbol.value + rela.symbol.baseaddr + rela.addend
            return val.to_bytes(self.addrsz, "big")
        elif rela.type == R_PPC_JUMP_SLOT:
            # The PowerPC64 JUMP_SLOT relocation is much more complicated.
            # It actually fills in three 64-bit values:
            #
            # 1. The actual function address
            # 2. The TOC base address; serves a similar purpose to the Global Pointer in MIPS.
            # 3. An "environment" pointer, not used by C.
            #
            # We're currently correctly filling in 1.
            # Entry 2. is the address of the GOT section of the containing binary + 0x8000.
            #
            # This requires a reference to the containing binary,
            # which I have no idea how to pass.
            raise NotImplementedError("R_PPC64_JUMP_SLOT not implemented")
        else:
            return super()._compute_value(rela)



from .defs import PlatformDef, RegisterAliasDef, RegisterDef
from .platforms import ABI, Architecture, Byteorder, Platform

__all__ = [
    "Platform",
    "Architecture",
    "Byteorder",
    "ABI",
    "PlatformDef",
    "RegisterDef",
    "RegisterAliasDef",
]



from ....posix.libgen import Basename, Dirname
from ..systemv import AArch64SysVModel


class AArch64SysVBasename(Basename, AArch64SysVModel):
    pass


class AArch64SysVDirname(Dirname, AArch64SysVModel):
    pass


__all__ = [
    "AArch64SysVBasename",
    "AArch64SysVDirname",
]



import abc
import typing

import lief

from ...... import exceptions, platforms, utils
from .....cpus import CPU
from .....state import FixedRegister

ET_CORE = 4


class PrStatus:
    """Class for extracting process information from a core dump

    The prstatus struct is arch-specific.
    Lief can interpret them for a small number of architectures,
    but definitely not all supported by SmallWorld.
    """

    @property
    @abc.abstractmethod
    def architecture(self) -> platforms.Architecture:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def byteorder(self) -> platforms.Byteorder:
        raise NotImplementedError()

    def __init__(self, elf: lief.ELF.Binary):
        assert elf.header.file_type.value == ET_CORE

        self.platform = platforms.Platform(self.architecture, self.byteorder)
        self.platdef = platforms.PlatformDef.for_platform(self.platform)

        self._registers: typing.Dict[str, int] = dict()

        # Find the PrStatus note
        prstatus: typing.Optional[lief.ELF.CorePrStatus] = None
        for note in elf.notes:
            if isinstance(note, lief.ELF.CorePrStatus):
                prstatus = note
                break

        if prstatus is None:
            # Didn't find one.
            raise exceptions.ConfigurationError("ELF core file has no PrStatus note")

        self._parse_prstatus(prstatus)

    @property
    @abc.abstractmethod
    def pr_regs_off(self) -> int:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def pr_regs_size(self) -> int:
        raise NotImplementedError()

    @property
    def register_coords(
        self,
    ) -> typing.List[typing.Tuple[typing.Optional[str], int, int]]:
        raise NotImplementedError()

    def _parse_prstatus(self, prstatus: lief.ELF.CorePrStatus) -> None:
        # Load the registers from prstatus.
        #
        # All prstatus structs have the same layout:
        # a single field called 'pr_regs' that contains a list of registers.
        # The only trick is knowng the offset of that field
        # and the offset and size of each register within it.

        # Get byteorder literals
        byteorder: typing.Literal["little", "big"]
        if self.byteorder is platforms.Byteorder.LITTLE:
            byteorder = "little"
        elif self.byteorder is platforms.Byteorder.BIG:
            # NOTE: Core dumps generated on a little-endian machine appear to come out little-endian.
            # This may be a problem in the future
            byteorder = "little"
        else:
            raise exceptions.ConfigurationError(
                f"Can't decode byteorder {self.byteorder}"
            )

        # Extract the raw bytes
        data = prstatus.description.tobytes()

        # Extract the registers from the struct
        for name, reg_off, reg_size in self.register_coords:
            if name is None:
                # Register we don't handle, or padding.
                continue
            reg_start = self.pr_regs_off + reg_off
            reg_end = self.pr_regs_off + reg_off + reg_size

            reg_bytes = data[reg_start:reg_end]
            if (
                self.platdef.address_size == 8
                and self.byteorder is platforms.Byteorder.BIG
            ):
                # 64-bit registers are somehow presented middle-endian
                # I have no idea.
                reg_bytes = reg_bytes[4:8] + reg_bytes[0:4]

            self._registers[name] = int.from_bytes(reg_bytes, byteorder)

    def populate_cpu(self, cpu: CPU) -> None:
        for reg, value in self._registers.items():
            if not hasattr(cpu, reg):
                raise exceptions.ConfigurationError(
                    f"PrStatus register {reg} not in CPU for platform {self.platform}"
                )
            field = getattr(cpu, reg)
            if not isinstance(field, FixedRegister):
                field.set(value)

    @classmethod
    def for_platform(cls, platform: platforms.Platform, elf: lief.ELF.Binary):
        try:
            return utils.find_subclass(
                cls,
                lambda x: x.architecture == platform.architecture
                and x.byteorder == platform.byteorder,
                elf,
            )
        except Exception as e:
            raise ValueError(f"No PrStatus class for {platform}") from e



from ....c99 import (
    Abort,
    Abs,
    Atexit,
    Atof,
    Atoi,
    Atol,
    Atoll,
    Calloc,
    Div,
    Exit,
    Free,
    Getenv,
    LAbs,
    LDiv,
    LLAbs,
    LLDiv,
    Malloc,
    Mblen,
    Mbstowcs,
    Mbtowc,
    QSort,
    Rand,
    Realloc,
    Srand,
    System,
    Wcstombs,
    Wctomb,
)
from ..systemv import MIPS64SysVModel


class MIPS64SysVAbs(Abs, MIPS64SysVModel):
    pass


class MIPS64SysVLAbs(LAbs, MIPS64SysVModel):
    pass


class MIPS64SysVLLAbs(LLAbs, MIPS64SysVModel):
    pass


class MIPS64SysVAtof(Atof, MIPS64SysVModel):
    pass


class MIPS64SysVAtoi(Atoi, MIPS64SysVModel):
    pass


class MIPS64SysVAtol(Atol, MIPS64SysVModel):
    pass


class MIPS64SysVAtoll(Atoll, MIPS64SysVModel):
    pass


class MIPS64SysVCalloc(Calloc, MIPS64SysVModel):
    pass


class MIPS64SysVDiv(Div, MIPS64SysVModel):
    pass


class MIPS64SysVLDiv(LDiv, MIPS64SysVModel):
    pass


class MIPS64SysVLLDiv(LLDiv, MIPS64SysVModel):
    pass


class MIPS64SysVExit(Exit, MIPS64SysVModel):
    pass


class MIPS64SysVFree(Free, MIPS64SysVModel):
    pass


class MIPS64SysVMalloc(Malloc, MIPS64SysVModel):
    pass


class MIPS64SysVQSort(QSort, MIPS64SysVModel):
    pass


class MIPS64SysVRand(Rand, MIPS64SysVModel):
    pass


class MIPS64SysVRealloc(Realloc, MIPS64SysVModel):
    pass


class MIPS64SysVSrand(Srand, MIPS64SysVModel):
    pass


class MIPS64SysVAbort(Abort, MIPS64SysVModel):
    pass


class MIPS64SysVAtexit(Atexit, MIPS64SysVModel):
    pass


class MIPS64SysVGetenv(Getenv, MIPS64SysVModel):
    pass


class MIPS64SysVMblen(Mblen, MIPS64SysVModel):
    pass


class MIPS64SysVMbstowcs(Mbstowcs, MIPS64SysVModel):
    pass


class MIPS64SysVMbtowc(Mbtowc, MIPS64SysVModel):
    pass


class MIPS64SysVSystem(System, MIPS64SysVModel):
    pass


class MIPS64SysVWcstombs(Wcstombs, MIPS64SysVModel):
    pass


class MIPS64SysVWctomb(Wctomb, MIPS64SysVModel):
    pass


__all__ = [
    "MIPS64SysVAbs",
    "MIPS64SysVLAbs",
    "MIPS64SysVLLAbs",
    "MIPS64SysVAtof",
    "MIPS64SysVAtoi",
    "MIPS64SysVAtol",
    "MIPS64SysVAtoll",
    "MIPS64SysVCalloc",
    "MIPS64SysVDiv",
    "MIPS64SysVLDiv",
    "MIPS64SysVLLDiv",
    "MIPS64SysVExit",
    "MIPS64SysVFree",
    "MIPS64SysVMalloc",
    "MIPS64SysVQSort",
    "MIPS64SysVRand",
    "MIPS64SysVRealloc",
    "MIPS64SysVSrand",
    "MIPS64SysVAbort",
    "MIPS64SysVAtexit",
    "MIPS64SysVGetenv",
    "MIPS64SysVMblen",
    "MIPS64SysVMbstowcs",
    "MIPS64SysVMbtowc",
    "MIPS64SysVSystem",
    "MIPS64SysVWcstombs",
    "MIPS64SysVWctomb",
]



import angr

from ....emulators.angr.exploration import (
    BoundedExplorationMixin,
    TerminationExplorationMixin,
)
from ....emulators.angr.memory import TrackerMemoryMixin
from .divergence import DivergenceExplorationMixin, DivergenceMemoryMixin
from .model import ModelMemoryMixin
from .typedefs import TypeDefPlugin


class NWBTMemoryPlugin(  # type: ignore[misc]
    DivergenceMemoryMixin,
    TrackerMemoryMixin,
    ModelMemoryMixin,
    angr.storage.DefaultMemory,
):
    pass


class NWBTExplorationTechnique(
    TerminationExplorationMixin,
    BoundedExplorationMixin,
    DivergenceExplorationMixin,
    angr.exploration_techniques.suggestions.Suggestions,
):
    pass


def configure_nwbt_plugins(emu):
    """Configure NWBT analysis plugins.

    This creates a new plugin preset that overrides
    angr's default symbolic memory plugin.
    This preset is passed to the entry state constructor,
    and can't be changed afterward.
    Thus, this needs to get called in a preinit callback.
    """
    preset = angr.SimState._presets["default"].copy()
    preset.add_default_plugin("sym_memory", NWBTMemoryPlugin)
    preset.add_default_plugin("typedefs", TypeDefPlugin)
    emu._plugin_preset = preset


def configure_nwbt_strategy(emu):
    """Configure NWBT analysis strategies

    This overrides the default angr exploration strategy.
    This needs to access the instantiated exploration manager,
    so it needs to get called in an init callback.
    """
    emu.mgr.use_technique(NWBTExplorationTechnique())
    # angr bug: states don't inherit plugin presets.
    #
    # Normally, they copy all plugins from their parents.
    # If you define a custom plugin and don't touch it,
    # it never gets initialized, and won't get inherited.
    #
    # If you try to touch that plugin on a successor,
    # it can't be initialized, since it doesn't have
    # your custom preset.
    emu.state.get_plugin("typedefs")



import logging

log = logging.getLogger(__name__)


class BoundedExplorationMixin:
    """
    Mixin forcing execution to obey our code bounds.
    """

    def step_state(self, simgr, state, **kwargs):
        if not state._ip.symbolic:
            ip = state._ip.concrete_value
            (r, found) = state.scratch.memory_map.find_closest_range(ip)
            if not found:
                return dict()
            (_, stop) = r
            size = stop - ip
            if not state.scratch.bounds.is_empty():
                (r, found) = state.scratch.bounds.find_closest_range(ip)
                if not found:
                    return dict()
                (_, stop) = r
                size = min(size, stop - ip)

            kwargs["size"] = size
        return super().step_state(simgr, state, **kwargs)



import capstone

from .bsid import BSIDMemoryReferenceOperand
from .instructions import Instruction, MemoryReferenceOperand


class AArch64Instruction(Instruction):
    angr_arch = "AARCH64"
    cs_arch = capstone.CS_ARCH_ARM64
    cs_mode = capstone.CS_MODE_ARM

    def _memory_reference(self, operand) -> MemoryReferenceOperand:
        # TODO: AArch64 operands have no size; what should it be?
        # I suspect it's always the same; is it always 64-bit?
        return BSIDMemoryReferenceOperand(
            base=self._instruction.reg_name(operand.value.mem.base),
            index=self._instruction.reg_name(operand.value.mem.index),
            offset=operand.value.mem.disp,
            size=8,
        )



from ....c99 import (
    Asctime,
    Clock,
    Ctime,
    Difftime,
    Gmtime,
    Localtime,
    Mktime,
    Strftime,
    Time,
)
from ..systemv import I386SysVModel


class I386SysVTime(Time, I386SysVModel):
    pass


class I386SysVLocaltime(Localtime, I386SysVModel):
    pass


class I386SysVGmtime(Gmtime, I386SysVModel):
    pass


class I386SysVCtime(Ctime, I386SysVModel):
    pass


class I386SysVAsctime(Asctime, I386SysVModel):
    pass


class I386SysVStrftime(Strftime, I386SysVModel):
    pass


class I386SysVDifftime(Difftime, I386SysVModel):
    pass


class I386SysVMktime(Mktime, I386SysVModel):
    pass


class I386SysVClock(Clock, I386SysVModel):
    pass


__all__ = [
    "I386SysVTime",
    "I386SysVLocaltime",
    "I386SysVGmtime",
    "I386SysVCtime",
    "I386SysVAsctime",
    "I386SysVStrftime",
    "I386SysVDifftime",
    "I386SysVMktime",
    "I386SysVClock",
]



import typing

from ..platforms import Architecture, Byteorder
from .platformdef import PlatformDef, RegisterAliasDef, RegisterDef


class LoongArchPlatformDef(PlatformDef):
    byteorder = Byteorder.LITTLE

    # NOTE: Capstone does not yet support LoongArch
    capstone_arch = -1
    capstone_mode = -1

    pc_register = "pc"
    sp_register = "r3"

    conditional_branch_mnemonics = {
        "beq",
        "bne",
        "blt",
        "bge",
        "bltu",
        "bgeu",
        "beqz",
    }
    # LoongArch follows MIPS' philosophy.
    # Integer comparisons are built into the conditional branch instructions.
    # I'd bet there are conditional FPU instructions,
    # if I can find them.
    compare_mnemonics = set()

    @property
    def general_purpose_registers(self) -> typing.List[str]:
        # Special registers:
        # r0: Hard-wired to zero
        # r1: Return address
        # r2: TLS pointer
        # r3: Stack pointer
        # r21: Per-CPU pointer
        return ["r{i}" for i in range(4, 32) if i != 21]

    @property
    def registers(self) -> typing.Dict[str, RegisterDef]:
        return self._registers

    def __init__(self):
        self._registers = {
            # Program counter (not a real register)
            "pc": RegisterDef(name="pc", size=self.address_size),
            # Zero register
            "r0": RegisterDef(name="r0", size=self.address_size),
            "zero": RegisterAliasDef(
                name="zero", parent="r0", size=self.address_size, offset=0
            ),
            # Return address
            "r1": RegisterDef(name="r1", size=self.address_size),
            "ra": RegisterAliasDef(
                name="ra", parent="r1", size=self.address_size, offset=0
            ),
            # TLS pointer
            "r2": RegisterDef(name="r2", size=self.address_size),
            "tp": RegisterAliasDef(
                name="tp", parent="r2", size=self.address_size, offset=0
            ),
            # Stack pointer
            "r3": RegisterDef(name="r3", size=self.address_size),
            "sp": RegisterAliasDef(
                name="sp", parent="r3", size=self.address_size, offset=0
            ),
            # Arguments.
            # a0 and a1 are also the return registers
            "r4": RegisterDef(name="r4", size=self.address_size),
            "a0": RegisterAliasDef(
                name="a0", parent="r4", size=self.address_size, offset=0
            ),
            "v0": RegisterAliasDef(
                name="v0", parent="r4", size=self.address_size, offset=0
            ),
            "r5": RegisterDef(name="r5", size=self.address_size),
            "a1": RegisterAliasDef(
                name="a1", parent="r5", size=self.address_size, offset=0
            ),
            "v1": RegisterAliasDef(
                name="v1", parent="r5", size=self.address_size, offset=0
            ),
            "r6": RegisterDef(name="r6", size=self.address_size),
            "a2": RegisterAliasDef(
                name="a2", parent="r6", size=self.address_size, offset=0
            ),
            "r7": RegisterDef(name="r7", size=self.address_size),
            "a3": RegisterAliasDef(
                name="a3", parent="r7", size=self.address_size, offset=0
            ),
            "r8": RegisterDef(name="r8", size=self.address_size),
            "a4": RegisterAliasDef(
                name="a4", parent="r8", size=self.address_size, offset=0
            ),
            "r9": RegisterDef(name="r9", size=self.address_size),
            "a5": RegisterAliasDef(
                name="a5", parent="r9", size=self.address_size, offset=0
            ),
            "r10": RegisterDef(name="r10", size=self.address_size),
            "a6": RegisterAliasDef(
                name="a6", parent="r10", size=self.address_size, offset=0
            ),
            "r11": RegisterDef(name="r11", size=self.address_size),
            "a7": RegisterAliasDef(
                name="a7", parent="r11", size=self.address_size, offset=0
            ),
            # Temporary registers
            "r12": RegisterDef(name="r12", size=self.address_size),
            "t0": RegisterAliasDef(
                name="t0", parent="r12", size=self.address_size, offset=0
            ),
            "r13": RegisterDef(name="r13", size=self.address_size),
            "t1": RegisterAliasDef(
                name="t1", parent="r13", size=self.address_size, offset=0
            ),
            "r14": RegisterDef(name="r14", size=self.address_size),
            "t2": RegisterAliasDef(
                name="t2", parent="r14", size=self.address_size, offset=0
            ),
            "r15": RegisterDef(name="r15", size=self.address_size),
            "t3": RegisterAliasDef(
                name="t3", parent="r15", size=self.address_size, offset=0
            ),
            "r16": RegisterDef(name="r16", size=self.address_size),
            "t4": RegisterAliasDef(
                name="t4", parent="r16", size=self.address_size, offset=0
            ),
            "r17": RegisterDef(name="r17", size=self.address_size),
            "t5": RegisterAliasDef(
                name="t5", parent="r17", size=self.address_size, offset=0
            ),
            "r18": RegisterDef(name="r18", size=self.address_size),
            "t6": RegisterAliasDef(
                name="t6", parent="r18", size=self.address_size, offset=0
            ),
            "r19": RegisterDef(name="r19", size=self.address_size),
            "t7": RegisterAliasDef(
                name="t7", parent="r19", size=self.address_size, offset=0
            ),
            "r20": RegisterDef(name="r20", size=self.address_size),
            "t8": RegisterAliasDef(
                name="t8", parent="r20", size=self.address_size, offset=0
            ),
            # Per-CPU Base Address
            "r21": RegisterDef(name="r21", size=self.address_size),
            "u0": RegisterAliasDef(
                name="u0", parent="r21", size=self.address_size, offset=0
            ),
            # Frame Pointer
            "r22": RegisterDef(name="r22", size=self.address_size),
            "fp": RegisterAliasDef(
                name="fp", parent="r22", size=self.address_size, offset=0
            ),
            # Static registers
            "r23": RegisterDef(name="r23", size=self.address_size),
            "s0": RegisterAliasDef(
                name="s0", parent="r23", size=self.address_size, offset=0
            ),
            "r24": RegisterDef(name="r24", size=self.address_size),
            "s1": RegisterAliasDef(
                name="s1", parent="r24", size=self.address_size, offset=0
            ),
            "r25": RegisterDef(name="r25", size=self.address_size),
            "s2": RegisterAliasDef(
                name="s2", parent="r25", size=self.address_size, offset=0
            ),
            "r26": RegisterDef(name="r26", size=self.address_size),
            "s3": RegisterAliasDef(
                name="s3", parent="r26", size=self.address_size, offset=0
            ),
            "r27": RegisterDef(name="r27", size=self.address_size),
            "s4": RegisterAliasDef(
                name="s4", parent="r27", size=self.address_size, offset=0
            ),
            "r28": RegisterDef(name="r28", size=self.address_size),
            "s5": RegisterAliasDef(
                name="s5", parent="r28", size=self.address_size, offset=0
            ),
            "r29": RegisterDef(name="r29", size=self.address_size),
            "s6": RegisterAliasDef(
                name="s6", parent="r29", size=self.address_size, offset=0
            ),
            "r30": RegisterDef(name="r30", size=self.address_size),
            "s7": RegisterAliasDef(
                name="s7", parent="r30", size=self.address_size, offset=0
            ),
            "r31": RegisterDef(name="r31", size=self.address_size),
            "s8": RegisterAliasDef(
                name="s8", parent="r31", size=self.address_size, offset=0
            ),
            # Floating-point arguments.
            # fa0 and fa1 are also return values
            "f0": RegisterDef(name="f0", size=8),
            "fa0": RegisterAliasDef(name="fa0", parent="f0", size=8, offset=0),
            "f1": RegisterDef(name="f1", size=8),
            "fa1": RegisterAliasDef(name="fa1", parent="f1", size=8, offset=0),
            "f2": RegisterDef(name="f2", size=8),
            "fa2": RegisterAliasDef(name="fa2", parent="f2", size=8, offset=0),
            "f3": RegisterDef(name="f3", size=8),
            "fa3": RegisterAliasDef(name="fa3", parent="f3", size=8, offset=0),
            "f4": RegisterDef(name="f4", size=8),
            "fa4": RegisterAliasDef(name="fa4", parent="f4", size=8, offset=0),
            "f5": RegisterDef(name="f5", size=8),
            "fa5": RegisterAliasDef(name="fa5", parent="f5", size=8, offset=0),
            "f6": RegisterDef(name="f6", size=8),
            "fa6": RegisterAliasDef(name="fa6", parent="f6", size=8, offset=0),
            "f7": RegisterDef(name="f7", size=8),
            "fa7": RegisterAliasDef(name="fa7", parent="f7", size=8, offset=0),
            # Floating-point temporary registers
            "f8": RegisterDef(name="f8", size=8),
            "ft0": RegisterAliasDef(name="ft0", parent="f8", size=8, offset=0),
            "f9": RegisterDef(name="f9", size=8),
            "ft1": RegisterAliasDef(name="ft1", parent="f9", size=8, offset=0),
            "f10": RegisterDef(name="f10", size=8),
            "ft2": RegisterAliasDef(name="ft2", parent="f10", size=8, offset=0),
            "f11": RegisterDef(name="f11", size=8),
            "ft3": RegisterAliasDef(name="ft3", parent="f11", size=8, offset=0),
            "f12": RegisterDef(name="f12", size=8),
            "ft4": RegisterAliasDef(name="ft4", parent="f12", size=8, offset=0),
            "f13": RegisterDef(name="f13", size=8),
            "ft5": RegisterAliasDef(name="ft5", parent="f13", size=8, offset=0),
            "f14": RegisterDef(name="f14", size=8),
            "ft6": RegisterAliasDef(name="ft6", parent="f14", size=8, offset=0),
            "f15": RegisterDef(name="f15", size=8),
            "ft7": RegisterAliasDef(name="ft7", parent="f15", size=8, offset=0),
            "f16": RegisterDef(name="f16", size=8),
            "ft8": RegisterAliasDef(name="ft8", parent="f16", size=8, offset=0),
            "f17": RegisterDef(name="f17", size=8),
            "ft9": RegisterAliasDef(name="ft9", parent="f17", size=8, offset=0),
            "f18": RegisterDef(name="f18", size=8),
            "ft10": RegisterAliasDef(name="ft10", parent="f18", size=8, offset=0),
            "f19": RegisterDef(name="f19", size=8),
            "ft11": RegisterAliasDef(name="ft11", parent="f19", size=8, offset=0),
            "f20": RegisterDef(name="f20", size=8),
            "ft12": RegisterAliasDef(name="ft12", parent="f20", size=8, offset=0),
            "f21": RegisterDef(name="f21", size=8),
            "ft13": RegisterAliasDef(name="ft13", parent="f21", size=8, offset=0),
            "f22": RegisterDef(name="f22", size=8),
            "ft14": RegisterAliasDef(name="ft14", parent="f22", size=8, offset=0),
            "f23": RegisterDef(name="f23", size=8),
            "ft15": RegisterAliasDef(name="ft15", parent="f23", size=8, offset=0),
            # Floating-point static registers
            "f24": RegisterDef(name="f24", size=8),
            "fs0": RegisterAliasDef(name="fs0", parent="f24", size=8, offset=0),
            "f25": RegisterDef(name="f25", size=8),
            "fs1": RegisterAliasDef(name="fs1", parent="f25", size=8, offset=0),
            "f26": RegisterDef(name="f26", size=8),
            "fs2": RegisterAliasDef(name="fs2", parent="f26", size=8, offset=0),
            "f27": RegisterDef(name="f27", size=8),
            "fs3": RegisterAliasDef(name="fs3", parent="f27", size=8, offset=0),
            "f28": RegisterDef(name="f28", size=8),
            "fs4": RegisterAliasDef(name="fs4", parent="f28", size=8, offset=0),
            "f29": RegisterDef(name="f29", size=8),
            "fs5": RegisterAliasDef(name="fs5", parent="f29", size=8, offset=0),
            "f30": RegisterDef(name="f30", size=8),
            "fs6": RegisterAliasDef(name="fs6", parent="f30", size=8, offset=0),
            "f31": RegisterDef(name="f31", size=8),
            "fs7": RegisterAliasDef(name="fs7", parent="f31", size=8, offset=0),
        }


class LoongArch64(LoongArchPlatformDef):
    architecture = Architecture.LOONGARCH64
    address_size = 8


__all__ = ["LoongArch64"]



from .hinting import *  # noqa: F401, F403
from .hinting import __all__ as __hinting__
from .hints import *  # noqa: F401, F403
from .hints import __all__ as __hints__
from .utils import *  # noqa: F401, F403
from .utils import __all__ as __utils__

__all__ = __hinting__ + __hints__ + __utils__



from ....c99 import (
    Asctime,
    Clock,
    Ctime,
    Difftime,
    Gmtime,
    Localtime,
    Mktime,
    Strftime,
    Time,
)
from ..systemv import LoongArch64SysVModel


class LoongArch64SysVTime(Time, LoongArch64SysVModel):
    pass


class LoongArch64SysVLocaltime(Localtime, LoongArch64SysVModel):
    pass


class LoongArch64SysVGmtime(Gmtime, LoongArch64SysVModel):
    pass


class LoongArch64SysVCtime(Ctime, LoongArch64SysVModel):
    pass


class LoongArch64SysVAsctime(Asctime, LoongArch64SysVModel):
    pass


class LoongArch64SysVStrftime(Strftime, LoongArch64SysVModel):
    pass


class LoongArch64SysVDifftime(Difftime, LoongArch64SysVModel):
    pass


class LoongArch64SysVMktime(Mktime, LoongArch64SysVModel):
    pass


class LoongArch64SysVClock(Clock, LoongArch64SysVModel):
    pass


__all__ = [
    "LoongArch64SysVTime",
    "LoongArch64SysVLocaltime",
    "LoongArch64SysVGmtime",
    "LoongArch64SysVCtime",
    "LoongArch64SysVAsctime",
    "LoongArch64SysVStrftime",
    "LoongArch64SysVDifftime",
    "LoongArch64SysVMktime",
    "LoongArch64SysVClock",
]



from __future__ import annotations

import abc
import typing

import claripy

from .. import platforms, utils


class Emulator(utils.MetadataMixin, metaclass=abc.ABCMeta):
    """An emulation environment.

    Arguments:
        platform: Platform metadata for emulation.
    """

    def __init__(self, platform: platforms.Platform):
        self.platform: platforms.Platform = platform
        super().__init__()
        self._bounds: utils.RangeCollection = utils.RangeCollection()
        self._exit_points: typing.Set[int] = set()
        """Configured platform metadata."""

    @abc.abstractmethod
    def read_register_content(self, name: str) -> int:
        """Read the content of a register.

        Arguments:
            name: The name of the register.

        Returns:
            The register's content.

        Raises:
            SymbolicValueError: If the register contains a non-collapsible symbolic value
        """

        return 0

    def read_register_symbolic(self, name: str) -> claripy.ast.bv.BV:
        """Read the content of a register, allowing symbolic output.

        If the implementation of read_register_content()
        raises SymbolicValueError, this must be implemented.

        Arguments:
            name: The name of the register

        Returns:
            The register's content as a z3 expression
        """
        raise NotImplementedError("This emulator does not support symbolic values")

    def read_register_type(self, name: str) -> typing.Optional[typing.Any]:
        """Read the type of a register.

        Note:
            Implementing this behavior is optional and not necessarily
            supported by all Emulators.

        Arguments:
            name: The name of the register.

        Returns:
            The register's type
        """

        return None

    def read_register_label(self, name: str) -> typing.Optional[str]:
        """Read the label of a register.

        Note:
            Implementing this behavior is optional and not necessarily
            supported by all Emulators.

        Arguments:
            name: The name of the register.

        Returns:
            The register's label.
        """

        return None

    def read_register(self, name: str) -> int:
        """A helper to read the content of a register.

        Arguments:
            name: The name of the register.

        Returns:
            The register's content.
        """

        return self.read_register_content(name)

    @abc.abstractmethod
    def write_register_content(
        self, name: str, content: typing.Union[None, int, claripy.ast.bv.BV]
    ) -> None:
        """Write some content to a register.

        Arguments:
            name: The name of the register to write.
            content: The content to write.

        Raises:
            SymbolicValueError: If content is a bitvector expression, and this emulator doesn't support them.
        """

        pass

    def write_register_type(
        self, name: str, type: typing.Optional[typing.Any] = None
    ) -> None:
        """Write the type of a register.

        Note:
            Implementing this behavior is optional and not necessarily
            supported by all Emulators.

        Arguments:
            name: The name of the register.
            type: The type of the register to write. To unset the register
                type, this may be set to None.
        """

        pass

    def write_register_label(
        self, name: str, label: typing.Optional[str] = None
    ) -> None:
        """Write the label of a register.

        Note:
            Implementing this behavior is optional and not necessarily
            supported by all Emulators.

        Arguments:
            name: The name of the register.
            label: The label of the register to write. To unset the register
                label, this may be set to None.
        """

        pass

    def write_register(
        self, name: str, content: typing.Union[None, int, claripy.ast.bv.BV]
    ) -> None:
        """A helper to write the content of a register.

        Arguments:
            name: The name of the register.
            content: The content to write.

        Raises:
            TypeError: If content is a BV, and this emulator cannot handle bitvector expressions
        """

        return self.write_register_content(name, content)

    @abc.abstractmethod
    def read_memory_content(self, address: int, size: int) -> bytes:
        """Read memory content from a specific address.

        Arguments:
            address: The address of the memory region.
            size: The size of the memory region.

        Returns:
            `size` bytes read from `address`

        Raises:
            SymbolicValueError:  If the memory range contains a non-collapsible symbolic value
        """

        return b""

    def read_memory_symbolic(self, address: int, size: int) -> claripy.ast.bv.BV:
        """Read memory content from a specific address as a symbolic expression.

        If the implementation of read_register_content()
        raises SymbolicValueError, this must be implemented.

        Arguments:
            address: The address of the memory region.
            size: The size of the memory region.

        Returns:
            A z3 expression of `size * 8` bits read from `address`
        """
        raise NotImplementedError("This emulator does not support symbolic values")

    def read_memory_type(self, address: int, size: int) -> typing.Optional[typing.Any]:
        """Read memory type from a specific address.

        Note:
            Implementing this behavior is optional and not necessarily
            supported by all Emulators.

        Arguments:
            address: The address of the memory region.
            size: The size of the memory region.

        Returns:
            The memory region's type.
        """

        return None

    def read_memory_label(self, address: int, size: int) -> typing.Optional[str]:
        """Read memory label from a specific address.

        Note:
            Implementing this behavior is optional and not necessarily
            supported by all Emulators.

        Arguments:
            address: The address of the memory region.
            size: The size of the memory region.

        Returns:
            The memory region's label.
        """

        return None

    def read_memory(self, address: int, size: int) -> bytes:
        """A helper to read memory content from a specific address.

        Arguments:
            address: The address of the memory region.
            size: The size of the memory region.

        Returns:
            `size` bytes read from `address`.

        Raises:
            SymbolicValueError:  If the memory range contains a non-collapsible symbolic value
        """

        return self.read_memory_content(address, size)

    @abc.abstractmethod
    def map_memory(self, address: int, size: int) -> None:
        """Map memory of a given size.

        If the requested allocation overlaps with existing regions,
        this will fill in the gaps.

        Arguments:
            address: The requested address of the allocation.
            size: The size of the allocation.
        """

        pass

    @abc.abstractmethod
    def get_memory_map(self) -> typing.List[typing.Tuple[int, int]]:
        """Retrieve the memory map as understood by the emulator.

        Returns:
            The list of tuples (start, end) of the mapped segments
        """

        return []

    @abc.abstractmethod
    def write_memory_content(
        self, address: int, content: typing.Union[bytes, claripy.ast.bv.BV]
    ) -> None:
        """Write content to memory at a specific address.

        Note:
            Written memory should already be mapped by some call to
            `map_memory()`.

        Arguments:
            address: The address of the memory region.
            content: The content to write.

        Raises:
            TypeError: If content is a BV, and this emulator can't handle them
        """

        pass

    def write_memory_type(
        self, address: int, size: int, type: typing.Optional[typing.Any] = None
    ) -> None:
        """Set the type of memory at a specific address.

        Note:
            Implementing this behavior is optional and not necessarily
            supported by all Emulators.

        Arguments:
            address: The address of the memory region.
            size: The size of the memory region
            type: The type of the register to write. To unset the register
                type, this may be set to None.
        """

        pass

    def write_memory_label(
        self, address: int, size: int, label: typing.Optional[str] = None
    ) -> None:
        """Set the label of memory at a specific address.

        Note:
            Implementing this behavior is optional and not necessarily
            supported by all Emulators.

        Arguments:
            address: The address of the memory region.
            size: The size of the memory region
            label: The label of the register to write. To unset the register
                label, this may be set to None.
        """

        pass

    def write_memory(
        self, address: int, content: typing.Union[bytes, claripy.ast.bv.BV]
    ) -> None:
        """A helper to write content to memory at a specific address.

        Note:
            Written memory should already be mapped by some call to
            `map_memory()`.

        Arguments:
            address: The address of the memory region.
            content: The content to write.

        Raises:
            SymbolicValueError: If `content` is a bitvector expression, and this emulator doesn't support them.
        """

        self.write_memory_content(address, content)

    def write_code(self, address: int, content: bytes) -> None:
        """Write executable memory at a specific address.

        Implementations can take advantage of this
        if they store code and memory differently.

        Otherwise, it should be identical to `write_memory()`.

        Note:
            Written memory should already be mapped by some call to
            `map_memory()`.

        Arguments:
            address: The address of the memory region.
            content: The content to write.

        """
        self.write_memory_content(address, content)

    def get_bounds(self) -> typing.List[typing.Tuple[int, int]]:
        """Get a list of all registered execution bounds.

        Returns:
            A list of registered execution bounds.
        """

        return list(self._bounds.ranges)

    def add_bound(self, start: int, end: int) -> None:
        """Add valid execution bounds.

        If execution leaves these bounds Emulators should raise
        EmulationBoundsError.

        If execution bounds are not specified, Emulators should allow execution
        anywhere.

        Arguments:
            start: The start address of a valid executable region.
            end: The end address of a valid executable region.
        """

        self._bounds.add_range((start, end))

    def remove_bound(self, start: int, end: int) -> None:
        self._bounds.remove_range((start, end))

    _exit_points: typing.Set[int] = set()

    def get_exit_points(self) -> typing.Set[int]:
        """Get a list of all registered exit points.

        Returns:
            A list of registered exit points.
        """

        return self._exit_points

    def add_exit_point(self, address: int) -> None:
        """Add an exit point.

        If execution reaches an exit point emulation should stop.

        Arguments:
            address: The address of the exit point.
        """

        self._exit_points.add(address)

    @abc.abstractmethod
    def step_instruction(self) -> None:
        """Single instruction step execution.

        Raises:
            EmulationBoundsError: if execution steps out of bounds.
        """

        pass

    @abc.abstractmethod
    def step_block(self) -> None:
        """Single block step execution.

        Raises:
            EmulationBoundsError: if execution steps out of bounds.
        """

        pass

    def step(self) -> None:
        """Helper for single instruction step execution.

        Raises:
            EmulationBoundsError: if execution steps out of bounds.
        """

        return self.step_instruction()

    @abc.abstractmethod
    def run(self) -> None:
        """Run execution indefinitely.

        Emulation should stop if an exit point is reached or execution leaves
        valid bounds.

        Raises:
            EmulationBoundsError: if execution steps out of bounds.
        """

        pass

    def get_thumb(self) -> bool:
        """For applicable platforms, checks if the CPU is in ARM or Thumb mode.

        Returns:
            True if in Thumb mode, otherwise False.

        Raises:
            ConfigurationError: if not using an ARM32 platform.
            NotImplementedError: if not using a supported emulator.
        """
        raise NotImplementedError("get_thumb() is not supported for this emulator")

    def set_thumb(self, enabled=True) -> None:
        """For applicable platforms, set the CPU to Thumb or ARM mode.

        Arguments:
            enabled: True for Thumb mode, False for ARM mode.

        Raises:
            ConfigurationError: if not using an ARM32 platform.
            NotImplementedError: if not using a supported emulator.
        """
        raise NotImplementedError("set_thumb() is not supported for this emulator")

    @abc.abstractmethod
    def __repr__(self) -> str:
        return ""


class InstructionHookable(metaclass=abc.ABCMeta):
    """An Emulator mixin that supports instruction hooking."""

    @abc.abstractmethod
    def hook_instruction(
        self, address: int, function: typing.Callable[[Emulator], None]
    ) -> None:
        """Hook a specific instruction by address.

        Arguments:
            address: The address of the hook.
            function: The function to execute when the address is reached.

        Example:
            The hook function looks like::

                def hook(emulator: Emulator) -> None:
                    ...
        """

        pass

    @abc.abstractmethod
    def unhook_instruction(self, address: int) -> None:
        """Unhook a specific instruction by address.

        Arguments:
            address: The address of the hook to remove.
        """

        pass

    @abc.abstractmethod
    def hook_instructions(self, function: typing.Callable[[Emulator], None]) -> None:
        pass

    @abc.abstractmethod
    def unhook_instructions(self) -> None:
        """Unhook all system interrupts."""

        pass


class FunctionHookable(metaclass=abc.ABCMeta):
    """An Emulator mixin that supports function hooking."""

    @abc.abstractmethod
    def hook_function(
        self, address: int, function: typing.Callable[[Emulator], None]
    ) -> None:
        """Hook a specific function by address.

        After the hook function is called, the Emulator will step out of the
        current function so that the hook essentially replaces a function call
        to the given address.

        Arguments:
            address: The address of the hook.
            function: The function to execute when the address is reached.

        Example:
            The hook function looks like::

                def hook(emulator: Emulator) -> None:
                    ...
        """

        pass

    @abc.abstractmethod
    def unhook_function(self, address: int) -> None:
        """Unhook a specific function by address.

        Arguments:
            address: The address of the hook to remove.
        """

        pass


class SyscallHookable(metaclass=abc.ABCMeta):
    """An emulator mixin that supports syscall hooking."""

    @abc.abstractmethod
    def hook_syscall(
        self, number: int, function: typing.Callable[[Emulator], None]
    ) -> None:
        """Hook a specific syscall by number.

        This hook will fire if emulation hits a platform-specific syscall instruction
        invoking the specified syscall number.

        Emulation will resume at the instruction after the syscall.
        To change this, the handler should modify the PC register in the emulator.

        Arguments:
            number: The syscall number to handle
            function: The handler function for this syscall
        """
        pass

    @abc.abstractmethod
    def unhook_syscall(self, number: int) -> None:
        """Unhook a specific syscall by number.

        Arguments:
            number: The syscall number to unhook
        """
        pass

    @abc.abstractmethod
    def hook_syscalls(self, function: typing.Callable[[Emulator, int], None]) -> None:
        """Hook all syscalls

        This hook will fire if emulation hits a platform-specific syscall instruction.

        Emulation will resume at the instruction after the syscall.
        To change this, the handler should modify the PC register in the emulator.

        Arguments:
            function: The handler function for all syscalls
        """
        pass

    @abc.abstractmethod
    def unhook_syscalls(self) -> None:
        """Unhook all syscalls"""
        pass


class MemoryReadHookable(metaclass=abc.ABCMeta):
    """An Emulator mixin that supports memory read hooking."""

    @abc.abstractmethod
    def hook_memory_read(
        self,
        start: int,
        end: int,
        function: typing.Callable[[Emulator, int, int, bytes], typing.Optional[bytes]],
    ) -> None:
        """Hook memory reads within a given range, handling concrete values

        Note that this will trigger for any read that overlaps the specified range,
        and will report all data read, not just the overlap.
        Hooks should expect to handle partial or oddly-sized reads.

        If the hook chooses to override the data being read,
        it must return data of the same size as the original read.

        Arguments:
            start: The start address of the memory range to hook.
            end: The end address of the memory range to hook.
            function: The function to execute when the memory region is read.

        Example:
            The hook function looks like::

                def hook(
                    emulator: Emulator,
                    address: int,
                    size: int,
                    content: bytes
                ) -> bytes:
                    # "content" represents the information the emulator would have fetched.
                    # The return value is what should be reported to the emulator.
                    # If return is "None", the original value of "content" will be reported.
                    ...

        """

        pass

    def hook_memory_read_symbolic(
        self,
        start: int,
        end: int,
        function: typing.Callable[
            [Emulator, int, int, claripy.ast.bv.BV], typing.Optional[claripy.ast.bv.BV]
        ],
    ) -> None:
        """Hook memory reads within a given range, handling symbolic values

        Note that this will trigger for any read that overlaps the specified range,
        and will report all data read, not just the overlap.
        Hooks should expect to handle partial or oddly-sized reads.

        If the hook chooses to override the data being read,
        it must return data of the same size as the original read.

        Arguments:
            start: The start address of the memory range to hook.
            end: The end address of the memory range to hook.
            function: The function to execute when the memory region is read.

        Raises:
            NotImplementedError: If this emulator doesn't support symbolic operations

        Example:
            The hook function looks like::

                def hook(
                    emulator: Emulator,
                    address: int,
                    size: int,
                    content: claripy.ast.bv.BV
                ) -> typing.Optional[claripy.ast.bv.BV]:
                    # "content" represents the information the emulator would have fetched.
                    # The return value is what should be reported to the emulator.
                    # If return is "None", the original value of "content" will be reported.
                    ...
        """
        raise NotImplementedError("This emulator doesn't support symbolic operations")

    @abc.abstractmethod
    def unhook_memory_read(self, start: int, end: int) -> None:
        """Unhook a specific memory region read by address range.

        Arguments:
            start: The start address of the memory read hook to remove.
            end: The end address of the memory read hook to remove.
        """

        pass

    @abc.abstractmethod
    def hook_memory_reads(
        self,
        function: typing.Callable[[Emulator, int, int, bytes], typing.Optional[bytes]],
    ) -> None:
        """Hook all memory reads, handling concrete values

        Arguments:
            function: The function to execute when the memory region is read.

        Example:
            The hook function looks like::

                def hook(
                    emulator: Emulator,
                    address: int,
                    size: int,
                    data: bytes
                ) -> typing.Optional[bytes]:
                    # "data" represents the information the emulator would have fetched.
                    # The return value is what should be reported to the emulator.
                    # If return is "None", the original value of "content" will be reported.
                    ...
        """
        pass

    def hook_memory_reads_symbolic(
        self,
        function: typing.Callable[
            [Emulator, int, int, claripy.ast.bv.BV], typing.Optional[claripy.ast.bv.BV]
        ],
    ) -> None:
        """Hook all memory reads, handling concrete values

        Arguments:
            function: The function to execute when the memory region is read.

        Example:
            The hook function looks like::

                def hook(
                    emulator: Emulator,
                    address: int,
                    size: int,
                    content: claripy.ast.bv.BV
                ) -> typing.Optional[claripy.ast.bv.BV]:
                    # "data" represents the data fetched by the emulator
                    # The return value is what should be reported to the guest
                    ...
        """
        raise NotImplementedError("This emulator doesn't support symbolic operations")

    @abc.abstractmethod
    def unhook_memory_reads(self) -> None:
        """Unhook any 'all reads' handlers"""

        pass


class MemoryWriteHookable(metaclass=abc.ABCMeta):
    """An Emulator mixin that supports memory write hooking."""

    @abc.abstractmethod
    def hook_memory_write(
        self,
        start: int,
        end: int,
        function: typing.Callable[[Emulator, int, int, bytes], None],
    ) -> None:
        """Hook memory writes within a given range, handling concrete values.

        Note that this will trigger for any write that overlaps the specified range,
        and will report all data written, not just the overlap.
        Hooks should expect to handle partial or oddly-sized writes.

        Arguments:
            start: The start address of the memory range to hook.
            end: The end address of the memory range to hook.
            function: The function to execute when the memory region is written.

        Example:
            The hook function looks like::

                def hook(emulator: Emulator, address: int, size: int, content: bytes) -> None:
                    # "content" is the value written by the guest.
                    # Hooks are responsible for completing the write to the emulator's state
                    ...
        """

        pass

    def hook_memory_write_symbolic(
        self,
        start: int,
        end: int,
        function: typing.Callable[[Emulator, int, int, claripy.ast.bv.BV], None],
    ) -> None:
        """Hook memory writes within a given range, handling symbolic values

        Note that this will trigger for any write that overlaps the specified range,
        and will report all data written, not just the overlap.
        Hooks should expect to handle partial or oddly-sized writes.

        Arguments:
            start: The start address of the memory range to hook.
            end: The end address of the memory range to hook.
            function: The function to execute when the memory region is written.

        Raises:
            NotImplementedError: If this emulator doesn't support symbolic operations

        Example:
            The hook function looks like::

                def hook(emulator: Emulator, address: int, size: int, content: claripy.ast.bv.BV) -> None:
                    # "content" is the value written by the guest.
                    # Hooks are responsible for completing the write to the emulator's state
                    ...
        """
        raise NotImplementedError("This emulator doesn't support symbolic operations")

    @abc.abstractmethod
    def unhook_memory_write(self, start: int, end: int) -> None:
        """Unhook a specific memory region write by address range.

        Arguments:
            start: The start address of the memory write hook to remove.
            end: The end address of the memory write hook to remove.
        """

        pass

    @abc.abstractmethod
    def hook_memory_writes(
        self, function: typing.Callable[[Emulator, int, int, bytes], None]
    ) -> None:
        """Hook all memory writes, handling concrete values.

        Arguments:
            function: The function to execute when the memory region is written.

        Example:
            The hook function looks like::

                def hook(emulator: Emulator, address: int, size: int, content: bytes) -> None:
                    # "content" is the value written by the guest.
                    # Hooks are responsible for completing the write to the emulator's state
                    ...
        """

        pass

    def hook_memory_writes_symbolic(
        self,
        function: typing.Callable[[Emulator, int, int, claripy.ast.bv.BV], None],
    ) -> None:
        """Hook all memory writes, handling symbolic values

        Arguments:
            function: The function to execute when the memory region is written.

        Raises:
            NotImplementedError: If this emulator doesn't support symbolic operations

        Example:
            The hook function looks like::

                def hook(emulator: Emulator, address: int, size: int, content: claripy.ast.bv.BV) -> None:
                    # "content" is the value written by the guest.
                    # Hooks are responsible for completing the write to the emulator's state
                    ...
        """
        raise NotImplementedError("This emulator doesn't support symbolic operations")

    @abc.abstractmethod
    def unhook_memory_writes(self) -> None:
        """Unhook any global memory write hook."""

        pass


class InterruptHookable(metaclass=abc.ABCMeta):
    """An Emulator mixin that supports interrupt hooking."""

    @abc.abstractmethod
    def hook_interrupts(self, function: typing.Callable[[Emulator, int], None]) -> None:
        """Hook any system interrupts.

        Arguments:
            function: The function to execute when an interrupt is triggered.

        Example:
            The hook function looks like::

                def hook(emulator: Emulator, interrupt: int) -> None:
                    ...
        """

        pass

    @abc.abstractmethod
    def unhook_interrupts(self) -> None:
        """Unhook all system interrupts."""

        pass

    @abc.abstractmethod
    def hook_interrupt(
        self, interrupt: int, function: typing.Callable[[Emulator], None]
    ) -> None:
        """Hook a specific system interrupt.

        Arguments:
            function: The function to execute when the interrupt is triggered.

        Example:
            The hook function looks like::

                def hook(emulator: Emulator) -> None:
                    ...
        """

        pass

    @abc.abstractmethod
    def unhook_interrupt(self, interupt: int) -> None:
        """Unhook a specific system interrupt.

        Arguments:
            interrupt: The interrupt to unhook.
        """

        pass


class ConstrainedEmulator:
    """Emulator that supports constraints

    It must also support some means of evaluating constraints,
    probably an SMT solver or similar.
    """

    @abc.abstractmethod
    def add_constraint(self, expr: claripy.ast.bool.Bool) -> None:
        """Add a constraint to the emulator

        A constraint is an expression that
        this emulator will use to limit the possible values
        of unbound variables.
        It will only consider execution states
        where all constraints can evaluate to True.

        Constraints must be Boolean expressions;
        the easiest form is the equality or inequality
        of two bitvector expressions.

        Arguments:
            expr: The constraint expression to add
        """
        pass

    @abc.abstractmethod
    def get_constraints(self) -> typing.List[claripy.ast.bool.Bool]:
        """Retrieve all constraints applied to this emulator.

        Returns:
            A list of constraint expressions
        """
        raise NotImplementedError("Abstract method")

    @abc.abstractmethod
    def satisfiable(
        self,
        extra_constraints: typing.List[claripy.ast.bool.Bool] = [],
    ) -> bool:
        """Check if the current set of constraints is satisfiable

        This checks if there's a way to assign variables
        such that all constraint expressions evaluate to "True".
        If not, the state can't exist as described.

        The emulator tracks its own set of constraints,
        added by the harness or built up durring execution.
        The caller can provide additional constraints
        for testing. These are not permanently added to the emulator.

        Arguments:
            extra_constraints:  A list of additional constraints to consider.

        Returns:
            True if the constraint system is satisfiable.  False otherwise.
        """
        return False

    @abc.abstractmethod
    def eval_atmost(self, expr: claripy.ast.bv.BV, most: int) -> typing.List[int]:
        """Find a maximum number of solutions to a bitvector expression

        This attempts to find concrete solutions to `expr`
        given the constraints on the emulator.

        It will return between 1 and `most` solutions, inclusive.
        It will raise exceptions if there are no solutions,
        or more than requested.

        Arguments:
            expr: The expression to evaluate
            most: The inclusive upper limit on solutions

        Returns:
            A list of integer solutions to `expr`

        Raises:
            UnsatError: If there are no solutions for `expr` given constraints
            SymbolicValueError: If there are more than `most` solutions for `expr` given constraints
        """
        raise NotImplementedError("Abstract method")

    @abc.abstractmethod
    def eval_atleast(self, expr: claripy.ast.bv.BV, least: int) -> typing.List[int]:
        """Find a minimum number of solutions to a bitvector expression

        This attempts to find concrete solutions to `expr`
        given the constraints on the emulator.

        It will return `least` solutions.
        It will raise an exception if there are fewer than `least` solutions possible.

        Arguments:
            expr: The expression to evaluate
            least: The number of solutions to retrieve

        Returns:
            A list of integer solutions to `expr`

        Raises:
            UnsatError: If there are no solutions for `expr` given constraints
            SymbolicValueError: If there are fewer than `least` solutions for `expr` given constraints
        """
        raise NotImplementedError("Abstract method")


__all__ = [
    "Emulator",
    "InstructionHookable",
    "FunctionHookable",
    "MemoryReadHookable",
    "MemoryWriteHookable",
    "InterruptHookable",
    "ConstrainedEmulator",
]



from . import cpus, memory, models
from .state import *  # noqa: F401, F403
from .state import __all__ as __state__
from .unstable import *  # noqa: F401, F403

__all__ = __state__ + ["cpus", "models", "memory"]



from __future__ import annotations

import abc
import collections
import copy
import ctypes
import logging as lg
import typing

import claripy

from .. import analyses, emulators, exceptions, logging, platforms, state, utils

logger = lg.getLogger(__name__)


class Stateful(metaclass=abc.ABCMeta):
    """An abstract class whose subclasses represent system state that can be applied to/loaded from an emulator."""

    @abc.abstractmethod
    def extract(self, emulator: emulators.Emulator) -> None:
        """Load state from an emulator.

        Arguments:
            emulator: The emulator from which to load
        """

        pass

    @abc.abstractmethod
    def apply(self, emulator: emulators.Emulator) -> None:
        """Apply state to an emulator.

        Arguments:
            emulator: The emulator to which state should applied.
        """

        pass

    def __hash__(self) -> int:
        return id(self)


class Value(metaclass=abc.ABCMeta):
    """An abstract class whose subclasses all have a tuple of content, type, and label.  Content is the value which must be convertable into bytes. The type is a ctype reprensenting the type of content. Label is a string that is a human label for the object. Any or all are optional."""

    def __init__(self: typing.Any) -> None:
        self._content: typing.Union[None, int, bytes, claripy.ast.bv.BV] = None
        self._type: typing.Optional[typing.Any] = None
        self._label: typing.Optional[str] = None

    @abc.abstractmethod
    def get_size(self) -> int:
        """Get the size of this value.

        Returns:
            The number of bytes this value should occupy in memory.
        """

        return 0

    def get_content(self) -> typing.Union[None, int, bytes, claripy.ast.bv.BV]:
        """Get the content of this value.

        Returns:
            The content of this value.
        """

        return self._content

    def set_content(
        self, content: typing.Union[None, int, bytes, claripy.ast.bv.BV]
    ) -> None:
        """Set the content of this value.

        Arguments:
            content: The content to which the value will be set.
        """

        self._content = content

    def get_type(self) -> typing.Optional[typing.Any]:
        """Get the type of this value.

        Returns:
            The type of this value.
        """

        return self._type

    def set_type(self, type: typing.Optional[typing.Any]) -> None:
        """Set the type of this value.

        Arguments:
            type: The type value to set.
        """

        self._type = type

    def get_label(self) -> typing.Optional[str]:
        """Get the label of this value.

        Returns:
            The label of this value.
        """

        return self._label

    def set_label(self, label: typing.Optional[str]) -> None:
        """Set the label of this value.

        Arguments:
            type: The label value to set.
        """

        self._label = label

    def get(self) -> typing.Union[None, int, bytes, claripy.ast.bv.BV]:
        """A helper to get the content of this value.

        Returns:
            The content of this value.
        """

        return self.get_content()

    def set(self, content: typing.Union[None, int, bytes, claripy.ast.bv.BV]) -> None:
        """A helper to set the content of this value.

        Arguments:
            content: The content value to set.
        """

        self.set_content(content)

    def to_symbolic(
        self, byteorder: platforms.Byteorder
    ) -> typing.Optional[claripy.ast.bv.BV]:
        """Convert this value into a symbolic expression

        For an unlabeled value, this will be a bit vector symbol containing the label.
        Otherwise, it will be a bit vector value containing the contents.

        Arguments:
            byteorder: The byte order to use in the conversion.

        Returns:
            Symbolic expression object, or None if both content and label are None
        """

        size = self.get_size()

        if self.get_label() is not None:
            # This has a label; assume we use it.
            label = self.get_label()
            if not isinstance(label, str):
                raise exceptions.ConfigurationError(
                    f"Cannot create a symbol from label of type {type(self._label)}; must be str"
                )

            # Bit vectors are bit vectors; multiply size in bytes by 8
            return claripy.BVS(label, size * 8, explicit_name=True)

        elif self._content is not None:
            content = self._content

            if isinstance(content, claripy.ast.bv.BV):
                # The content is already a symbolic expression
                return content

            if isinstance(content, int):
                # The content is an int; convert to bytes for universal handling
                if byteorder == platforms.Byteorder.BIG:
                    content = content.to_bytes(size, "big")
                else:
                    content = content.to_bytes(size, "little")

            if not isinstance(content, bytes) and not isinstance(content, bytearray):
                # The content is not something I know how to handle.
                raise exceptions.ConfigurationError(
                    f"Cannot create a symbol from content of type {type(self._content)}; "
                    "must be a claripy expression, int, bytes, or bytearray"
                )

            if size == 0:
                raise exceptions.ConfigurationError(
                    "Cannot create a bitvector of size zero"
                )

            if len(content) != size:
                raise exceptions.ConfigurationError(
                    f"Expected size {size}, but content has size {len(content)}"
                )

            return claripy.BVV(content)

        else:
            return None

    def __getstate__(self):
        # Override default pickling operation
        # The '_type' field is a ctypes class which won't be pickleable.
        # TODO: figure out how to pickle ctypes classes
        state = self.__dict__.copy()
        state["_type"] = None
        return state

    @abc.abstractmethod
    def to_bytes(self, byteorder: platforms.Byteorder) -> bytes:
        """Convert this value into a byte string.

        Arguments:
            byteorder: Byteorder for conversion to raw bytes.

        Returns:
            Bytes for this value with the given byteorder.
        """

        return b""

    @classmethod
    def from_ctypes(cls, ctype: typing.Any, label: str):
        """Load from an existing ctypes value.

        Arguements:
            ctype: The data in ctype form.

        Reutrns:
            The value constructed from ctypes data.
        """

        class CTypeValue(Value):
            def __init__(self, ctype: typing.Any, label: str):
                self._content = ctype
                self._label = label
                self._type = ctype.__class__

            def __getstate__(self):
                state = self.__dict__.copy()
                del state["_content"]
                del state["_type"]
                return state

            # def __setstate__(self, state):

            def get_size(self) -> int:
                return ctypes.sizeof(self._content)  # type: ignore

            def to_bytes(self, byteorder: platforms.Byteorder) -> bytes:
                return bytes(self._content)  # type: ignore

        return CTypeValue(ctype, label)


class SymbolicValue(Value):
    """A symbolic value

    This value is expressed as a bitvector expression,
    compatible with angr's internal state representation.
    If used with a concrete emulator, it defaults to zeroes.

    Arguments:
        size: The size of the region
        bv: Optional bitvector value
        type: Optional typedef information
        label: An optional metadata label
    """

    def __init__(
        self,
        size: int,
        bv: typing.Optional[claripy.ast.bv.BV],
        type: typing.Optional[typing.Any],
        label: typing.Optional[str],
    ):
        super().__init__()
        self._content = bv
        self._size = size
        self._type = type
        self._label = label

    def get_size(self) -> int:
        return self._size

    def to_bytes(self):
        return b"\0" * self._size


class IntegerValue(Value):
    """An integer value.

    This is useful for using python integers, but passing them to emulators that care about things like width. The type is derived based on the size.

    Arguments:
        size: The size of the integer. Must be 1, 2, 4, 8
        label: An optional metadata label
    """

    def __init__(
        self, integer: int, size: int, label: typing.Optional[str], signed: bool = True
    ) -> None:
        super().__init__()
        if size == 8:
            if signed:
                self._type = ctypes.c_int64
            else:
                self._type = ctypes.c_uint64
        elif size == 4:
            if signed:
                self._type = ctypes.c_int32
            else:
                self._type = ctypes.c_uint32
        elif size == 2:
            if signed:
                self._type = ctypes.c_int16
            else:
                self._type = ctypes.c_uint16
        elif size == 1:
            if signed:
                self._type = ctypes.c_int8
            else:
                self._type = ctypes.c_uint8
        else:
            raise NotImplementedError(f"{size}-bit integers are not yet implemented")
        self._content = integer
        self._label = label
        self._size = size

    def get_size(self) -> int:
        return self._size

    def to_bytes(self, byteorder: platforms.Byteorder) -> bytes:
        if self._content is None:
            raise ValueError("IntegerValue must have an integer value")
        if not isinstance(self._content, int):
            raise TypeError("IntegerValue is not an integer")
        value = self._content
        if value < 0:
            # Convert signed python into unsigned int containing 2s-compliment value.
            # Python's to_bytes() doesn't do this on its own.
            value = 2 ** (self._size * 8) + value
        if byteorder == platforms.Byteorder.LITTLE:
            return value.to_bytes(self._size, byteorder="little")
        elif byteorder == platforms.Byteorder.BIG:
            return value.to_bytes(self._size, byteorder="big")
        else:
            raise NotImplementedError("middle endian integers are not yet implemented")


class BytesValue(Value):
    """A bytes value.

    This is for representing a python bytes or bytearray as a value.

    Arguments:
        label: An optional metadata label
    """

    def __init__(
        self, content: typing.Union[bytes, bytearray], label: typing.Optional[str]
    ) -> None:
        super().__init__()
        self._content = bytes(content)
        self._label = label
        self._size = len(self._content)
        self._type = ctypes.c_ubyte * self._size

    def get_size(self) -> int:
        return self._size

    def to_bytes(self, byteorder: platforms.Byteorder) -> bytes:
        if self._content is None or not isinstance(self._content, bytes):
            raise ValueError("BytesValue must have a bytes value")
        return self._content


class Register(Value, Stateful):
    """An individual register.

    Arguments:
        name: The canonical name of the register.
        size: The size (in bytes) of the register.
    """

    def __init__(self, name: str, size: int = 4):
        super().__init__()

        self.name: str = name
        """Canonical name."""

        self.size = size
        """Register size in bytes."""

    def __str__(self):
        s = f"Reg({self.name},{self.size})="
        x = self.get_content()
        if x is None:
            s = s + "=None"
        elif isinstance(x, int):
            s = s + f"0x{x:x}"
        else:
            s = s + str(type(x))
        return s

    def set_content(self, content: typing.Union[None, int, bytes, claripy.ast.bv.BV]):
        if content is not None:
            if not isinstance(content, int) and not isinstance(
                content, claripy.ast.bv.BV
            ):
                raise TypeError(
                    f"Expected None, int, or claripy expression as content for Register {self.name}, got {type(content)}"
                )
            if isinstance(content, int) and content < 0:
                logger.warn(
                    "Converting content {hex(content)} of {self.name} to unsigned."
                )
                content = content + (2 ** (self.size * 8))
            if isinstance(content, claripy.ast.bv.BV):
                if content.size() != self.size * 8:
                    raise ValueError(
                        f"Claripy content had size {content.size()} bits, but expected {self.size * 8} bits"
                    )
        super().set_content(content)

    def get_size(self) -> int:
        return self.size

    def extract(self, emulator: emulators.Emulator) -> None:
        content: typing.Union[int, claripy.ast.bv.BV] = 0
        try:
            content = emulator.read_register_content(self.name)
            if content is not None:
                self.set_content(content)
        except exceptions.SymbolicValueError:
            content = emulator.read_register_symbolic(self.name)
            if content is not None:
                self.set_content(content)
        except exceptions.UnsupportedRegisterError:
            return

        type = emulator.read_register_type(self.name)
        if type is not None:
            self.set_type(type)

        try:
            label = emulator.read_register_label(self.name)
            if label is not None:
                self.set_label(label)
        except exceptions.SymbolicValueError:
            pass

    def apply(self, emulator: emulators.Emulator) -> None:
        content = self.get_content()
        if isinstance(content, bytes):
            raise TypeError("Register content cannot be bytes")
        if content is not None:
            emulator.write_register_content(self.name, content)
        if self.get_type() is not None:
            emulator.write_register_type(self.name, self.get_type())
        if self.get_label() is not None:
            emulator.write_register_label(self.name, self.get_label())

    def to_bytes(self, byteorder: platforms.Byteorder) -> bytes:
        value = self.get_content()

        if value is None:
            # Default to zeros if no value present.
            return b"\0" * self.size
        elif isinstance(value, claripy.ast.bv.BV):
            raise exceptions.SymbolicValueError(
                "Cannot convert a symbolic expression to bytes"
            )
        elif isinstance(value, bytes):
            # This never happens, but let's keep mypy happy
            return value
        elif byteorder == platforms.Byteorder.LITTLE:
            return value.to_bytes(self.size, byteorder="little")
        elif byteorder == platforms.Byteorder.BIG:
            return value.to_bytes(self.size, byteorder="big")
        else:
            raise ValueError(f"unsupported byteorder {byteorder}")


class RegisterAlias(Register):
    """An alias to a partial register.

    Arguments:
        name: The cannonical name of the register.
        reference: A register which this alias references.
        size: The size (in bytes) of the register.
        offset: The offset from the start of the register that this alias
            references.

    """

    def __init__(self, name: str, reference: Register, size: int = 4, offset: int = 0):
        super().__init__(name, size)

        self.reference: Register = reference
        """The register referenced by this alias."""

        self.offset: int = offset
        """'The offset into the referenced register."""

    @property
    def mask(self) -> int:
        mask = (1 << self.size * 8) - 1
        mask <<= self.offset * 8

        return mask

    def get_content(self) -> typing.Union[None, int, bytes, claripy.ast.bv.BV]:
        r = self.reference.get_content()
        if r is None:
            return r
        value = self.reference.get_content()
        if value is not None:
            if isinstance(value, int):
                value = value & self.mask
                value >>= self.offset * 8
            elif isinstance(value, claripy.ast.bv.BV):
                lo = self.offset * 8
                hi = lo + self.size * 8 - 1
                value = claripy.simplify(value[hi:lo])
            else:
                raise TypeError(f"Unexpected register content {type(value)}")
        return value

    def set_content(
        self, content: typing.Union[None, int, bytes, claripy.ast.bv.BV]
    ) -> None:
        if content is not None:
            if isinstance(content, int):
                if content < 0:
                    logger.warn(
                        f"Converting content {hex(content)} of {self.name} to unsigned."
                    )
                    content = content + (2 ** (self.size * 8))

            elif isinstance(content, claripy.ast.bv.BV):
                # Bitvectors can only interoperate with bitvectors of the same size.
                content = claripy.ZeroExt(
                    self.reference.size * 8 - self.size * 8, content
                )

            else:
                raise TypeError(
                    f"Can only accept None, int, or BV, not {type(content)}"
                )

            value = self.reference.get_content()
            if value is None:
                value = 0
            if isinstance(value, bytes):
                raise TypeError("Value should not be bytes")

            # mypy completely loses the plot trying to determine type for content
            content = content << (self.offset * 8)  # type: ignore[operator]
            value = (value & ~self.mask) | content
            if isinstance(value, claripy.ast.bv.BV):
                value = claripy.simplify(value)

            self.reference.set_content(value)

    def get_type(self) -> typing.Optional[typing.Any]:
        return self.reference.get_type()

    def set_type(self, type: typing.Optional[typing.Any]) -> None:
        self.reference.set_type(type)

    def get_label(self) -> typing.Optional[str]:
        return self.reference.get_label()

    def set_label(self, label: typing.Optional[str]) -> None:
        self.reference.set_label(label)

    def extract(self, emulator: emulators.Emulator) -> None:
        pass

    def apply(self, emulator: emulators.Emulator) -> None:
        pass


class FixedRegister(Register):
    """A Register that holds a fixed value, and should not be set.

    A number of ISAs have registers hard-wired to zero.
    It would be nice for the harness author to not have
    to remember to set it each time.

    Arguments:
        name:   Name of the register
        size:   Size of the register in bytes
        value:  Fixed value of the register
    """

    def __init__(self, name, size=4, value=0):
        super().__init__(name, size=size)
        super().set_content(value)

    def set_content(self, value: typing.Optional[typing.Any]):
        raise exceptions.ConfigurationError(
            "Register {self.name} is fixed; it cannot be set"
        )

    def get_label(self) -> typing.Optional[str]:
        # Fixed registers cannot have labels
        return None

    def set_label(self, label: typing.Optional[str]) -> None:
        # Fixed registers cannot have labels
        pass

    def get_type(self) -> typing.Optional[typing.Any]:
        # Fixed registers do not have types
        return None

    def set_type(self, type: typing.Optional[typing.Any]) -> None:
        # Fixed registers do not have types
        pass

    def extract(self, emulator: emulators.Emulator) -> None:
        # Don't bother extracting content
        pass

    def apply(self, emulator: emulators.Emulator) -> None:
        try:
            super().apply(emulator)
        except exceptions.UnsupportedRegisterError:
            # If the register isn't supported, that's okay
            pass


class StatefulSet(Stateful, collections.abc.MutableSet):
    """A set that holds stateful objects. Applying or extracting the set performs the action of every member of the set."""

    def __init__(self):
        super().__init__()
        self._contents = set()

    def extract(self, emulator: emulators.Emulator) -> None:
        for stateful in self:
            stateful.extract(emulator)

    def apply(self, emulator: emulators.Emulator) -> None:
        for stateful in self:
            logger.debug(
                f"applying state {stateful} of type {type(stateful)} to emulator {emulator}"
            )
            stateful.apply(emulator)

    def __contains__(self, item):
        return item in self._contents

    def __iter__(self):
        return self._contents.__iter__()

    def __len__(self):
        return len(self._contents)

    def add(self, item):
        self._contents.add(item)

    def discard(self, item):
        self._contents.discard(item)

    def members(self, type):
        return set(filter(lambda x: isinstance(x, type), self._contents))


class Machine(StatefulSet):
    """A container for all state needed to begin or resume emulation or
    analysis), including CPU with register values, code, raw memory or
    even stack and heap memory.

    Machines have exit points which are instruction addresses that
    when hit by an emulator will cause it to stop before any side
    effects of that instruction are applied. Similarly, machines have
    bounds which are address ranges. When an address outside of the
    range is hit by the emulator that will cause it to stop before any
    side effects are applied. Note that the start of the range is
    included in the range, but the end is not.

    """

    def __init__(self):
        super().__init__()
        self._bounds = utils.RangeCollection()
        self._exit_points = set()
        self._constraints = list()

    def add_exit_point(self, address: int):
        """Add an exit point to the machine.

        Arguments:
            address: The address to exit on

        """
        self._exit_points.add(address)

    def get_exit_points(self) -> typing.Set[int]:
        """Gets the set of exit points for a machine.

        Returns:
            The set of exit point addresses.
        """
        return self._exit_points

    def add_bound(self, start: int, end: int):
        """Adds a bound to the machine

        Arguments:
            start: the start address of the bound (included in the bound)
            end: the end address of the bound (excluded in the bound)
        """
        self._bounds.add_range((start, end))

    def get_bounds(self) -> typing.List[typing.Tuple[int, int]]:
        """Gets a list of bounds for the machine.


        Returns:
            The list of bounds.
        """
        return list(self._bounds.ranges)

    def add_constraint(self, expr: claripy.ast.bool.Bool) -> None:
        """Add a constraint to the environment

        A constraint is an expression that
        some emulators can use to limit the possible values
        of unbound variables.
        They will only consider execution states
        where all constraints can evaluate to True.

        Constraints must be Boolean expressions;
        the easiest form is the equality or inequality
        of two bitvector expressions.

        You can get the variable representing
        a labeled Value via its `to_symbolic()` method.
        Note that Values with both a label and content
        already have a constraint binding the label's
        variable to the content.

        Arguments:
            expr: The constraint expression to add
        """
        if not isinstance(expr, claripy.ast.bool.Bool):
            raise TypeError(f"expr is a {type(expr)}, not a Boolean expression")

        self._constraints.append(expr)

    def get_constraints(self) -> typing.List[claripy.ast.bool.Bool]:
        """Retrieve all constraints applied to this machine.

        Returns:
            A list of constraint expressions
        """
        return list(self._constraints)

    def apply(self, emulator: emulators.Emulator) -> None:
        for address in self._exit_points:
            emulator.add_exit_point(address)
        for start, end in self.get_bounds():
            emulator.add_bound(start, end)

        if isinstance(emulator, emulators.ConstrainedEmulator):
            for expr in self._constraints:
                emulator.add_constraint(expr)

        return super().apply(emulator)

    def extract(self, emulator: emulators.Emulator) -> None:
        self._exit_points = emulator.get_exit_points()
        for start, end in emulator.get_bounds():
            self.add_bound(start, end)

        if isinstance(emulator, emulators.ConstrainedEmulator):
            self._constraints = emulator.get_constraints()

        return super().extract(emulator)

    def emulate(self, emulator: emulators.Emulator) -> Machine:
        """Emulate this machine with the given emulator.

        Arguments:
            emulator: An emulator instance on which this machine state should
                run.

        Returns:
            The final system state after emulation.
        """

        self.apply(emulator)

        try:
            emulator.run()
        except exceptions.EmulationBounds:
            pass

        machine_copy = copy.deepcopy(self)
        machine_copy.extract(emulator)

        return machine_copy

    def analyze(self, analysis: analyses.Analysis) -> None:
        """Run the given analysis on this machine.

        Arguments:
            analysis: The analysis to run.
        """

        analysis.run(self)

    def step(
        self, emulator: emulators.Emulator
    ) -> typing.Generator[Machine, None, None]:
        """This is a generator that single steps the machine each time it is called.

        Yields:
            A new machine that is the previous one single stepped forward.
        """
        self.apply(emulator)

        while True:
            try:
                emulator.step()
                machine_copy = copy.deepcopy(self)
                machine_copy.extract(emulator)
                yield machine_copy

            except exceptions.EmulationStop:
                print(
                    "emulation complete; encountered exit point or went out of bounds"
                )
                break
        return None

    def fuzz(
        self,
        emulator: emulators.Emulator,
        input_callback: typing.Callable,
        crash_callback: typing.Optional[typing.Callable] = None,
        always_validate: bool = False,
        iterations: int = 1,
    ) -> None:
        """Fuzz the machine using unicornafl.

        Arguments:
            emulator: Currently, must be the unicorn emulator
            input_callback: A callback that applies an input to a machine
            input_file_path: The path of the input file AFL will mutate. If not given, we assume argv[1].
            crash_callback: An optional callback that is given the unicorn state and can decide whether or not to record it as a crash. (See unicornafl documentation for more info)
            always_validate: Whether to run the crash_callback on every run or only when unicorn returns an error.
            iterations: The number of iterations to run before forking a new child
        Returns:
            Bytes for this value with the given byteorder.
        """
        import argparse

        arg_parser = argparse.ArgumentParser(description="AFL Harness")
        arg_parser.add_argument(
            "input_file", type=str, help="File path AFL will mutate"
        )
        args = arg_parser.parse_args()
        self.fuzz_with_file(
            emulator,
            input_callback,
            args.input_file,
            crash_callback,
            always_validate,
            iterations,
        )

    def fuzz_with_file(
        self,
        emulator: emulators.Emulator,
        input_callback: typing.Callable,
        input_file_path: str,
        crash_callback: typing.Optional[typing.Callable] = None,
        always_validate: bool = False,
        iterations: int = 1,
    ) -> None:
        """Fuzz the machine using unicornafl.

        Arguments:
            emulator: Currently, must be the unicorn emulator
            input_callback: A callback that applies an input to a machine
            input_file_path: The path of the input file AFL will mutate. If not given, we assume argv[1].
            crash_callback: An optional callback that is given the unicorn state and can decide whether or not to record it as a crash. (See unicornafl documentation for more info)
            always_validate: Whether to run the crash_callback on every run or only when unicorn returns an error.
            iterations: The number of iterations to run before forking a new child
        Returns:
            Bytes for this value with the given byteorder.
        """
        try:
            import unicornafl
        except ImportError:
            raise RuntimeError(
                "missing `unicornafl` - afl++ must be installed manually from source"
            )

        if not isinstance(emulator, emulators.UnicornEmulator):
            raise RuntimeError("you must use a unicorn emulator to fuzz")

        self.apply(emulator)

        unicornafl.uc_afl_fuzz(
            uc=emulator.engine,
            input_file=input_file_path,
            place_input_callback=input_callback,
            exits=emulator.get_exit_points(),
            validate_crash_callback=crash_callback,
            always_validate=always_validate,
            persistent_iters=iterations,
        )

    def get_cpus(self):
        """Gets a list of :class:`~smallworld.state.cpus.cpu.CPU` attached to this machine.

        Returns:
            A list of objects that subclass :class:`~smallworld.state.cpus.cpu.CPU` attached to this machin.
        """
        return [i for i in self if issubclass(type(i), state.cpus.cpu.CPU)]

    def get_elfs(self):
        return [i for i in self if issubclass(type(i), state.memory.ElfExecutable)]

    def get_platforms(self):
        """Gets a set of platforms for the :class:`~smallworld.state.cpus.cpu.CPU` (s) attached to this machine.

        Returns:
            A set of platforms for the :class:`~smallworld.state.cpus.cpu.CPU` (s) attached to this machin.

        """
        return set([i.get_platform() for i in self.get_cpus()])

    def get_cpu(self):
        """Gets the :class:`~smallworld.state.cpus.cpu.CPU` attached to this machine. Throws an exception if the machine has more than one.

        Raises:
            ConfigurationError: if the machine has more than one :class:`~smallworld.state.cpus.cpu.CPU`

        Returns:
            The :class:`~smallworld.state.cpus.cpu.CPU` attached to this machine.
        """
        cpus = self.get_cpus()
        if len(cpus) != 1:
            raise exceptions.ConfigurationError("You have more than one CPU")
        return cpus[0]

    def get_elf(self):
        elfs = self.get_elfs()
        if len(elfs) != 1:
            raise exceptions.ConfigurationError("You have more than one ELF")
        return elfs[0]

    def get_platform(self):
        """Gets the platform of the :class:`~smallworld.state.cpus.cpu.CPU` (s) attached to this machine. Throws an exception if the machine has more than one.

        Raises:
            ConfigurationError: if the machine has more than one platform

        Returns:
            The platform for the :class:`~smallworld.state.cpus.cpu.CPU` (s) attached to this machine.
        """
        platforms = self.get_platforms()
        if len(platforms) != 1:
            raise exceptions.ConfigurationError("You have more than one platform")
        return platforms.pop()

    def read_memory(self, address: int, size: int) -> typing.Optional[bytes]:
        """Read bytes out of memory if available.

        Arguments:
            address: start address of read.
            size: number of bytes to read.

        Returns:
            the bytes read, or None if unavailable.
        """
        for m in self:
            if issubclass(type(m), state.memory.Memory):
                for po, v in m.items():
                    if m.address + po <= address <= m.address + po + v.get_size():
                        c = m[po].get()
                        o = address - (m.address + po)
                        return c[o : o + size]
        return None


__all__ = [
    "Stateful",
    "Value",
    "IntegerValue",
    "BytesValue",
    "SymbolicValue",
    "Register",
    "RegisterAlias",
    "FixedRegister",
    "Machine",
]



import logging

from angr.storage.memory_mixins.memory_mixin import MemoryMixin

from ..utils import reg_name_from_offset

log = logging.getLogger(__name__)


class TrackerMemoryMixin(MemoryMixin):
    """Memory mixin for tracking used data.

    Printing every AMD64 register is a pain,
    and finding used memory in angr is a pain.
    This tracks which parts of memory are in use,
    mostly to avoid overprinting.

    NOTE: This class needs to go in a pretty specific place in a memory mixin hierarchy.

    It needs to go above anything that overrides _default_value(),
    or else it will pick up defaulting addresses before they are defined,
    possibly resulting in recursive calls if you try to use pp().

    Conversely, it should go below anything that alters
    what is actually getting passed to the default memory mixins,
    such as changes to address concretization.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dirty = dict()
        self.dirty_addrs = set()

    @MemoryMixin.memo
    def copy(self, memo):
        o = super().copy(memo)
        o.dirty = dict()
        o.dirty.update(self.dirty)
        o.dirty_addrs = set()
        o.dirty_addrs.update(self.dirty_addrs)
        return o

    def _track_memory(self, addr, size):
        """Mark a memory range as 'in-use'"""
        # TODO: Make this more efficient
        # Detecting overlapping ranges is a pain.
        # This implementation is easy, but its memory usage
        # and runtime cost will explode violently
        # if large amounts of memory are in use.
        track = False
        for i in range(0, size):
            if not track and addr + i not in self.dirty_addrs:
                track = True
                self.dirty[addr] = size
            self.dirty_addrs.add(addr + i)
        if track:
            log.debug(f"Tracking {addr:x}, {size}")

    def _default_value(self, addr, size, **kwargs):
        out = super()._default_value(addr, size, **kwargs)
        self._track_memory(addr, size)
        return out

    def _store_one_addr(
        self, concrete_addr, data, trivial, addr, condition, size, **kwargs
    ):
        out = super()._store_one_addr(
            concrete_addr, data, trivial, addr, condition, size, **kwargs
        )
        self._track_memory(concrete_addr, size)
        return out

    def pp(self, log):
        addrs = list(self.dirty.keys())
        addrs.sort()
        for addr in addrs:
            code = False
            size = self.dirty[addr]
            if self.id == "reg":
                name = reg_name_from_offset(self.state.arch, addr, size)
            else:
                name = f"0x{addr:x}"
            if code:
                log(f"\t{name}: <code [{hex(size)} bytes]>")
            else:
                val = self.load(addr, size, disable_actions=True)
                log(f"\t{name}: {val}")

    def create_hint(self):
        if self.id == "reg":
            return {
                reg_name_from_offset(self.state.arch, addr, self.dirty[addr]): str(
                    self.load(addr, self.dirty[addr], disable_actions=True)
                )
                for addr in list(self.dirty.keys())
            }
        else:
            return {
                addr: str(
                    self.load(
                        addr, self.dirty[addr], disable_actions=True, inspect=False
                    )
                )
                for addr in list(self.dirty.keys())
            }



from smallworld.state.state import Machine

from ..hinting import BranchesHint, CoverageFrontierHint, Hint, TraceExecutionHint
from .analysis import Analysis


class CoverageFrontier(Analysis):
    name = "coverage_frontier"
    description = "deduce coverage frontier with respect to a set of traces"
    version = "0.0.1"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.edges = {}
        self.branches = set([])
        self.hinter.register(TraceExecutionHint, self.monitor_branches)

    def monitor_branches(self, hint: Hint) -> None:
        # examine execution trace and collect any branch information
        if isinstance(hint, TraceExecutionHint):
            lb = None
            for element in hint.trace:
                if lb is not None:
                    if lb.pc not in self.edges:
                        self.edges[lb.pc] = set([])
                    self.edges[lb.pc].add(element.pc)
                    lb = None
                if element.branch:
                    # instructions that are branches
                    self.branches.add(element.pc)
                    lb = element

    def run(self, machine: Machine) -> None:
        self.hinter.send(
            BranchesHint(message="Branches encountered", branches=list(self.branches))
        )
        # analyze to get coverage frontier,
        # i.e., set of half-covered  conditionals
        coverage_frontier = set([])
        for branch in self.branches:
            if (len(self.edges[branch])) == 1:
                coverage_frontier.add(branch)
        self.hinter.send(
            CoverageFrontierHint(
                message="Coverage frontier (half-covered conditionals) for a set of execution traces",
                coverage_frontier=list(coverage_frontier),
            )
        )



from .libgen import *  # noqa: F401, F403
from .libgen import __all__ as __libgen__
from .signal import *  # noqa: F401, F403
from .signal import __all__ as __signal__

__all__ = __libgen__ + __signal__



import typing

import capstone

from ..platforms import Architecture, Byteorder
from .platformdef import PlatformDef, RegisterAliasDef, RegisterDef


class ARMPlatformDef(PlatformDef):
    """Base class for ARM 32-bit platform definitions

    All ARM CPUs share the same basic registers,
    but there are at least two dimensions of difference:
    M-series vs A/R series, and FPU model.
    """

    byteorder = Byteorder.LITTLE

    address_size = 4

    capstone_arch = capstone.CS_ARCH_ARM
    capstone_mode = capstone.CS_MODE_ARM

    conditional_branch_mnemonics = {
        # Conditional branch
        "beq",
        "bne",
        "bcs",
        "bhs",
        "bcc",
        "blo",
        "bmi",
        "bpl",
        "bvs",
        "bvc",
        "bhi",
        "bls",
        "bge",
        "blt",
        "bgt",
        "ble",
        # Compare and branch on zero/non-zero
        "cbz",
        "cbnz",
    }

    pc_register = "pc"
    sp_register = "sp"

    # NOTE: r9, r10, r11, and r12 technically have special purposes,
    # but they're optional.
    general_purpose_registers = [f"r{i}" for i in range(0, 13)]

    @property
    def registers(self) -> typing.Dict[str, RegisterDef]:
        return self._registers

    @property
    def compare_mnemonics(self) -> typing.Set[str]:
        return self._compare_mnemonics

    def __init__(self):
        super().__init__()
        self._compare_mnemonics = {
            # Integer comparison
            "cmp",
            "cmn",
            "tst",
            # Conditional integer comparison
            "cmpeq",
            "cmpne",
            "cmpcs",
            "cmphs",
            "cmpcc",
            "cmplo",
            "cmpmi",
            "cmppl",
            "cmpvs",
            "cmpvc",
            "cmphi",
            "cmpls",
            "cmpge",
            "cmplt",
            "cmpgt",
            "cmple",
            "cmneq",
            "cmnne",
            "cmncs",
            "cmnhs",
            "cmncc",
            "cmnlo",
            "cmnmi",
            "cmnpl",
            "cmnvs",
            "cmnvc",
            "cmnhi",
            "cmnls",
            "cmnge",
            "cmnlt",
            "cmngt",
            "cmnle",
            "tsteq",
            "tstne",
            "tstcs",
            "tsths",
            "tstcc",
            "tstlo",
            "tstmi",
            "tstpl",
            "tstvs",
            "tstvc",
            "tsthi",
            "tstls",
            "tstge",
            "tstlt",
            "tstgt",
            "tstle",
        }
        self._registers = {
            # *** General-purpose registers ***
            "r0": RegisterDef(name="r0", size=4),
            "r1": RegisterDef(name="r1", size=4),
            "r2": RegisterDef(name="r2", size=4),
            "r3": RegisterDef(name="r3", size=4),
            "r4": RegisterDef(name="r4", size=4),
            "r5": RegisterDef(name="r5", size=4),
            "r6": RegisterDef(name="r6", size=4),
            "r7": RegisterDef(name="r7", size=4),
            "r8": RegisterDef(name="r8", size=4),
            # r9 doubles as the Static base pointer
            "r9": RegisterDef(name="r9", size=4),
            "sb": RegisterAliasDef(name="sb", parent="r9", size=4, offset=0),
            # r10 doubles as the Stack Limit pointer
            "r10": RegisterDef(name="r10", size=4),
            "sl": RegisterAliasDef(name="sl", parent="r10", size=4, offset=0),
            # r11 doubles as the Frame Pointer, if desired.
            "r11": RegisterDef(name="r11", size=4),
            "fp": RegisterAliasDef(name="fp", parent="r11", size=4, offset=0),
            # r12 doubles as the Intra-call scratch register
            "r12": RegisterDef(name="r12", size=4),
            "ip": RegisterAliasDef(name="ip", parent="r12", size=4, offset=0),
            # sp is technically also r13, but is never aliased as such
            "sp": RegisterDef(name="sp", size=4),
            # lr is technically also r14, but is never aliased as such
            "lr": RegisterDef(name="lr", size=4),
            # pc is technically also r15, but is never aliased as such
            "pc": RegisterDef(name="pc", size=4),
        }


class ARMPlatformMixinM:
    """Abstract class for M-series CPUs.

    The M-series CPUs are built for embedded applications,
    and have an extremely simplified privilege model,
    and don't often have an MMU.
    """

    def __init__(self):
        super().__init__()
        self._registers |= {
            # *** Special Registers ***
            # Program Status Register
            # NOTE: PSR can be accessed through several masked aliases.
            # These are read-only, so I'm not including them.
            # - apsr: Just the condition flags
            # - ipsr: Just exception information
            # - epsr: Just execution state info
            # - iapsr: apsr | ipsr
            # - eapsr: apsr | epsr
            # - iepsr: ipsr | epsr
            # - xpsr: apsr | ipsr | epsr
            "psr": RegisterDef(name="psr", size=4),
            # Exception Mask Register
            "primask": RegisterDef(name="primask", size=4),
            # Base Priority Mask Register
            "basepri": RegisterDef(name="basepri", size=4),
            # Fault Mask Register
            "faultmask": RegisterDef(name="faultmask", size=4),
            # Control register; includes a lot of flags.
            "control": RegisterDef(name="control", size=4),
            # *** Stack Pointer Bank ***
            # sp is actually an alias to one of these two.
            # Exactly which one depends on a bit in control.
            # Emulators that care should be careful when loading state.
            # Main Stack Pointer
            "msp": RegisterDef(name="msp", size=4),
            # Process Stack Pointer
            "psp": RegisterDef(name="psp", size=4),
        }


class ARMPlatformMixinRA:
    """Abstract class for R- or A-series CPUs.

    The A- series supports full application-style multi-tasking,
    and thus needs a much more complicated privileged model
    than the M-series.

    The R- series uses the same privilege model as the A- series,
    but it's designed for real-time operations.
    The major differences are in the MMU semantics, which don't impact registers.
    """

    def __init__(self):
        super().__init__()
        self._registers |= {
            # *** Special Registers ***
            # Current Program Status Register
            # NOTE: CPSR can be accessed through several masked aliases.
            # These are read-only, so I'm not including them.
            # - isetstate: Just includes instruction set control bits
            # - itstate: Just includes state bits for Thumb IT instruction
            "cpsr": RegisterDef(name="cpsr", size=4),
            # Saved Program Status Register
            "spsr": RegisterDef(name="spsr", size=4),
            # *** Register Banks ***
            # sp, lr, and spsr are actually aliases to one of these.
            # Which one they reference depends on execution mode.
            # Emulators that care should be careful when loading state.
            # NOTE: Use User-mode copies of registers unless the mode has its own.
            # NOTE: The user-mode bank is only used if you're in a different privilege mode.
            # User-mode Stack Pointer
            "sp_usr": RegisterDef(name="sp_usr", size=4),
            # User-mode Link Register
            "lr_usr": RegisterDef(name="lr_usr", size=4),
            # User-mode r8
            "r8_usr": RegisterDef(name="r8_usr", size=4),
            # User-mode r9
            "r9_usr": RegisterDef(name="r9_usr", size=4),
            # User-mode r10
            "r10_usr": RegisterDef(name="r10_usr", size=4),
            # User-mode r11
            "r11_usr": RegisterDef(name="r11_usr", size=4),
            # User-mode r12
            "r12_usr": RegisterDef(name="r12_usr", size=4),
            # Hypervisor Stack Pointer
            "sp_hyp": RegisterDef(name="sp_hyp", size=4),
            # Hypervisor Saved PSR
            "spsr_hyp": RegisterDef(name="spsr_hyp", size=4),
            # Hypervisor Exception Link Register
            # NOTE: This isn't so much banked, as it only exists in hypervisor mode.
            "elr_hyp": RegisterDef(name="elr_hyp", size=4),
            # Supervisor Stack Pointer
            "sp_svc": RegisterDef(name="sp_svc", size=4),
            # Supervisor Link Register
            "lr_svc": RegisterDef(name="lr_svc", size=4),
            # Supervisor Saved PSR
            "spsr_svc": RegisterDef(name="spsr_svc", size=4),
            # Abort-state Stack Pointer
            "sp_abt": RegisterDef(name="sp_abt", size=4),
            # Abort-state Link Register
            "lr_abt": RegisterDef(name="lr_abt", size=4),
            # Abort-state Saved PSR
            "spsr_abt": RegisterDef(name="spsr_abt", size=4),
            # Undefined-mode Stack Pointer
            "sp_und": RegisterDef(name="sp_und", size=4),
            # Undefined-mode Link Register
            "lr_und": RegisterDef(name="lr_und", size=4),
            # Undefined-mode Saved PSR
            "spsr_und": RegisterDef(name="spsr_und", size=4),
            # Monitor-mode Stack Pointer
            "sp_mon": RegisterDef(name="sp_mon", size=4),
            # Monitor-mode Link Register
            "lr_mon": RegisterDef(name="lr_mon", size=4),
            # Monitor-mode Saved PSR
            "spsr_mon": RegisterDef(name="spsr_mon", size=4),
            # IRQ-mode Stack Pointer
            "sp_irq": RegisterDef(name="sp_irq", size=4),
            # IRQ-mode Link Register
            "lr_irq": RegisterDef(name="lr_irq", size=4),
            # IRQ-mode Saved PSR
            "spsr_irq": RegisterDef(name="spsr_irq", size=4),
            # FIQ-mode Stack Pointer
            "sp_fiq": RegisterDef(name="sp_fiq", size=4),
            # FIQ-mode Link Register
            "lr_fiq": RegisterDef(name="lr_fiq", size=4),
            # FIQ-mode Saved PSR
            "spsr_fiq": RegisterDef(name="spsr_fiq", size=4),
            # FIQ-mode r8
            "r8_fiq": RegisterDef(name="r8_fiq", size=4),
            # FIQ-mode r9
            "r9_fiq": RegisterDef(name="r9_fiq", size=4),
            # FIQ-mode r10
            "r10_fiq": RegisterDef(name="r10_fiq", size=4),
            # FIQ-mode r11
            "r11_fiq": RegisterDef(name="r11_fiq", size=4),
            # FIQ-mode r12
            "r12_fiq": RegisterDef(name="r12_fiq", size=4),
        }


class ARMPlatformMixinFPEL:
    """Mixin for little-endian ARM CPUs with FP extensions

    This is a simpler floating-point extension
    which offers 64-bit scalar operations
    """

    def __init__(self):
        super().__init__()
        # TODO: What do the FP floating-point instructions look like?
        # I only have docs for VFP.
        self._registers |= {
            # *** Floating point control registers ***
            # Floating-point Status and Control Register
            "fpscr": RegisterDef(name="fpscr", size=4),
            # Floating-point Exception Control Register
            "fpexc": RegisterDef(name="fpexc", size=4),
            # Floating-point System ID Register
            "fpsid": RegisterDef(name="fpsid", size=4),
            # Media and VFP Feature Register 0
            "mvfr0": RegisterDef(name="mvfr0", size=4),
            # Media and VFP Feature Register 1
            "mvfr1": RegisterDef(name="mvfr1", size=4),
            # *** Floating point registers ***
            "d0": RegisterDef(name="d0", size=8),
            "s0": RegisterAliasDef(name="s0", parent="d0", size=4, offset=0),
            "s1": RegisterAliasDef(name="s1", parent="d0", size=4, offset=4),
            "d1": RegisterDef(name="d1", size=8),
            "s2": RegisterAliasDef(name="s2", parent="d1", size=4, offset=0),
            "s3": RegisterAliasDef(name="s3", parent="d1", size=4, offset=4),
            "d2": RegisterDef(name="d2", size=8),
            "s4": RegisterAliasDef(name="s4", parent="d2", size=4, offset=0),
            "s5": RegisterAliasDef(name="s5", parent="d2", size=4, offset=4),
            "d3": RegisterDef(name="d3", size=8),
            "s6": RegisterAliasDef(name="s6", parent="d3", size=4, offset=0),
            "s7": RegisterAliasDef(name="s7", parent="d3", size=4, offset=4),
            "d4": RegisterDef(name="d4", size=8),
            "s8": RegisterAliasDef(name="s8", parent="d4", size=4, offset=0),
            "s9": RegisterAliasDef(name="s9", parent="d4", size=4, offset=4),
            "d5": RegisterDef(name="d5", size=8),
            "s10": RegisterAliasDef(name="s10", parent="d5", size=4, offset=0),
            "s11": RegisterAliasDef(name="s11", parent="d5", size=4, offset=4),
            "d6": RegisterDef(name="d6", size=8),
            "s12": RegisterAliasDef(name="s12", parent="d6", size=4, offset=0),
            "s13": RegisterAliasDef(name="s13", parent="d6", size=4, offset=4),
            "d7": RegisterDef(name="d7", size=8),
            "s14": RegisterAliasDef(name="s14", parent="d7", size=4, offset=0),
            "s15": RegisterAliasDef(name="s15", parent="d7", size=4, offset=4),
            "d8": RegisterDef(name="d8", size=8),
            "s16": RegisterAliasDef(name="s16", parent="d8", size=4, offset=0),
            "s17": RegisterAliasDef(name="s17", parent="d8", size=4, offset=4),
            "d9": RegisterDef(name="d9", size=8),
            "s18": RegisterAliasDef(name="s18", parent="d9", size=4, offset=0),
            "s19": RegisterAliasDef(name="s19", parent="d9", size=4, offset=4),
            "d10": RegisterDef(name="d10", size=8),
            "s20": RegisterAliasDef(name="s20", parent="d10", size=4, offset=0),
            "s21": RegisterAliasDef(name="s21", parent="d10", size=4, offset=4),
            "d11": RegisterDef(name="d11", size=8),
            "s22": RegisterAliasDef(name="s22", parent="d11", size=4, offset=0),
            "s23": RegisterAliasDef(name="s23", parent="d11", size=4, offset=4),
            "d12": RegisterDef(name="d12", size=8),
            "s24": RegisterAliasDef(name="s24", parent="d12", size=4, offset=0),
            "s25": RegisterAliasDef(name="s25", parent="d12", size=4, offset=4),
            "d13": RegisterDef(name="d13", size=8),
            "s26": RegisterAliasDef(name="s26", parent="d13", size=4, offset=0),
            "s27": RegisterAliasDef(name="s27", parent="d13", size=4, offset=4),
            "d14": RegisterDef(name="d14", size=8),
            "s28": RegisterAliasDef(name="s28", parent="d14", size=4, offset=0),
            "s29": RegisterAliasDef(name="s29", parent="d14", size=4, offset=4),
            "d15": RegisterDef(name="d15", size=8),
            "s30": RegisterAliasDef(name="s30", parent="d15", size=4, offset=0),
            "s31": RegisterAliasDef(name="s31", parent="d15", size=4, offset=4),
        }


class ARMPlatformMixinVFPEL:
    """Mixin for little-endian ARM CPUs with VFP/NEON extensions

    This is a newer floating-point extension
    which supports up to 128-bit scalar and SIMD vector operations.

    VFP and NEON are always optional extensions;
    the two can exist independently, and VFP can support 16 or 32 double registers.
    This is the maximal set of registers, assuming both are supported.
    """

    def __init__(self):
        super().__init__()
        # TODO: Add the floating-point conditionals
        self._registers |= {
            # *** Floating-point Control Registers ***
            # Floating-point Status and Control Register
            "fpscr": RegisterDef(name="fpscr", size=4),
            # Floating-point Exception Control Register
            "fpexc": RegisterDef(name="fpexc", size=4),
            # Floating-point System ID Register
            "fpsid": RegisterDef(name="fpsid", size=4),
            # Media and VFP Feature Register 0
            "mvfr0": RegisterDef(name="mvfr0", size=4),
            # Media and VFP Feature Register 1
            "mvfr1": RegisterDef(name="mvfr1", size=4),
            # *** Floating-point Registers ****
            "q0": RegisterDef(name="q0", size=16),
            "d0": RegisterAliasDef(name="d0", parent="q0", size=8, offset=0),
            "s0": RegisterAliasDef(name="s0", parent="q0", size=4, offset=0),
            "s1": RegisterAliasDef(name="s1", parent="q0", size=4, offset=4),
            "d1": RegisterAliasDef(name="d1", parent="q0", size=8, offset=8),
            "s2": RegisterAliasDef(name="s2", parent="q0", size=4, offset=8),
            "s3": RegisterAliasDef(name="s3", parent="q0", size=4, offset=12),
            "q1": RegisterDef(name="q1", size=16),
            "d2": RegisterAliasDef(name="d2", parent="q1", size=8, offset=0),
            "s4": RegisterAliasDef(name="s4", parent="q1", size=4, offset=0),
            "s5": RegisterAliasDef(name="s5", parent="q1", size=4, offset=4),
            "d3": RegisterAliasDef(name="d3", parent="q1", size=8, offset=8),
            "s6": RegisterAliasDef(name="s6", parent="q1", size=4, offset=8),
            "s7": RegisterAliasDef(name="s7", parent="q1", size=4, offset=12),
            "q2": RegisterDef(name="q2", size=16),
            "d4": RegisterAliasDef(name="d4", parent="q2", size=8, offset=0),
            "s8": RegisterAliasDef(name="s8", parent="q2", size=4, offset=0),
            "s9": RegisterAliasDef(name="s9", parent="q2", size=4, offset=4),
            "d5": RegisterAliasDef(name="d5", parent="q2", size=8, offset=8),
            "s10": RegisterAliasDef(name="s10", parent="q2", size=4, offset=8),
            "s11": RegisterAliasDef(name="s11", parent="q2", size=4, offset=12),
            "q3": RegisterDef(name="q3", size=16),
            "d6": RegisterAliasDef(name="d6", parent="q3", size=8, offset=0),
            "s12": RegisterAliasDef(name="s12", parent="q3", size=4, offset=0),
            "s13": RegisterAliasDef(name="s13", parent="q3", size=4, offset=4),
            "d7": RegisterAliasDef(name="d7", parent="q3", size=8, offset=8),
            "s14": RegisterAliasDef(name="s14", parent="q3", size=4, offset=8),
            "s15": RegisterAliasDef(name="s15", parent="q3", size=4, offset=12),
            "q4": RegisterDef(name="q4", size=16),
            "d8": RegisterAliasDef(name="d8", parent="q4", size=8, offset=0),
            "s16": RegisterAliasDef(name="s16", parent="q4", size=4, offset=0),
            "s17": RegisterAliasDef(name="s17", parent="q4", size=4, offset=4),
            "d9": RegisterAliasDef(name="d9", parent="q4", size=8, offset=8),
            "s18": RegisterAliasDef(name="s18", parent="q4", size=4, offset=8),
            "s19": RegisterAliasDef(name="s19", parent="q4", size=4, offset=12),
            "q5": RegisterDef(name="q5", size=16),
            "d10": RegisterAliasDef(name="d10", parent="q5", size=8, offset=0),
            "s20": RegisterAliasDef(name="s20", parent="q5", size=4, offset=0),
            "s21": RegisterAliasDef(name="s21", parent="q5", size=4, offset=4),
            "d11": RegisterAliasDef(name="d11", parent="q5", size=8, offset=8),
            "s22": RegisterAliasDef(name="s22", parent="q5", size=4, offset=8),
            "s23": RegisterAliasDef(name="s23", parent="q5", size=4, offset=12),
            "q6": RegisterDef(name="q6", size=16),
            "d12": RegisterAliasDef(name="d12", parent="q6", size=8, offset=0),
            "s24": RegisterAliasDef(name="s24", parent="q6", size=4, offset=0),
            "s25": RegisterAliasDef(name="s25", parent="q6", size=4, offset=4),
            "d13": RegisterAliasDef(name="d13", parent="q6", size=8, offset=8),
            "s26": RegisterAliasDef(name="s26", parent="q6", size=4, offset=8),
            "s27": RegisterAliasDef(name="s27", parent="q6", size=4, offset=12),
            "q7": RegisterDef(name="q7", size=16),
            "d14": RegisterAliasDef(name="d14", parent="q7", size=8, offset=0),
            "s28": RegisterAliasDef(name="s28", parent="q7", size=4, offset=0),
            "s29": RegisterAliasDef(name="s29", parent="q7", size=4, offset=4),
            "d15": RegisterAliasDef(name="d15", parent="q7", size=8, offset=8),
            "s30": RegisterAliasDef(name="s30", parent="q7", size=4, offset=8),
            "s31": RegisterAliasDef(name="s31", parent="q7", size=4, offset=12),
            # NOTE: This isn't a typo; there are only 32 single-precision sX registers
            # This does mean that only half the VFP register space can be used
            # for single-precision arithmetic.
            "q8": RegisterDef(name="q8", size=16),
            "d16": RegisterAliasDef(name="d16", parent="q8", size=8, offset=0),
            "d17": RegisterAliasDef(name="d17", parent="q8", size=8, offset=8),
            "q9": RegisterDef(name="q9", size=16),
            "d18": RegisterAliasDef(name="d18", parent="q9", size=8, offset=0),
            "d19": RegisterAliasDef(name="d19", parent="q9", size=8, offset=8),
            "q10": RegisterDef(name="q10", size=16),
            "d20": RegisterAliasDef(name="d20", parent="q10", size=8, offset=0),
            "d21": RegisterAliasDef(name="d21", parent="q10", size=8, offset=8),
            "q11": RegisterDef(name="q11", size=16),
            "d22": RegisterAliasDef(name="d22", parent="q11", size=8, offset=0),
            "d23": RegisterAliasDef(name="d23", parent="q11", size=8, offset=8),
            "q12": RegisterDef(name="q12", size=16),
            "d24": RegisterAliasDef(name="d24", parent="q12", size=8, offset=0),
            "d25": RegisterAliasDef(name="d25", parent="q12", size=8, offset=8),
            "q13": RegisterDef(name="q13", size=16),
            "d26": RegisterAliasDef(name="d26", parent="q13", size=8, offset=0),
            "d27": RegisterAliasDef(name="d27", parent="q13", size=8, offset=8),
            "q14": RegisterDef(name="q14", size=16),
            "d28": RegisterAliasDef(name="d28", parent="q14", size=8, offset=0),
            "d29": RegisterAliasDef(name="d29", parent="q14", size=8, offset=8),
            "q15": RegisterDef(name="q15", size=16),
            "d30": RegisterAliasDef(name="d30", parent="q15", size=8, offset=0),
            "d31": RegisterAliasDef(name="d31", parent="q15", size=8, offset=8),
        }


class ARMv5T(ARMPlatformMixinM, ARMPlatformDef):
    """Platform definition for ARMv5t little-endian."""

    architecture = Architecture.ARM_V5T


class ARMv6M(ARMPlatformMixinFPEL, ARMPlatformMixinM, ARMPlatformDef):
    """Platform definition for ARMv6m little-endian."""

    architecture = Architecture.ARM_V6M


class ARMv6MThumb(ARMv6M):
    """Platform definition for ARMv6m little-endian in thumb mode.

    Thumb is annoying; most emulators have a very hard time
    switching between arm and thumb instructions,
    despite it being a key feature of many CPUs.
    """

    architecture = Architecture.ARM_V6M_THUMB


class ARMv7M(ARMPlatformMixinFPEL, ARMPlatformMixinM, ARMPlatformDef):
    """Platform definition for ARMv7m little-endian"""

    architecture = Architecture.ARM_V7M


class ARMv7R(ARMPlatformMixinVFPEL, ARMPlatformMixinRA, ARMPlatformDef):
    """Platform definition for ARMv7r little-endian"""

    architecture = Architecture.ARM_V7R


class ARMv7A(ARMPlatformMixinVFPEL, ARMPlatformMixinRA, ARMPlatformDef):
    """Platform definition for ARMv7a little-endian"""

    architecture = Architecture.ARM_V7A



import typing

from .... import platforms
from . import stack


class RISCV64Stack(stack.DescendingStack):
    """A stack for a RISCV 64-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.RISCV64, platforms.Byteorder.LITTLE
    )

    def get_alignment(self) -> int:
        return 16

    @classmethod
    def initialize_stack(cls, argv: typing.List[bytes], *args, **kwargs):
        raise NotImplementedError("Stack initialization not implemented for AArch64")



class RegisterState:
    def __init__(self, registers: dict, pc: int, sp: int, status: int, arch: str):
        self.arch = arch
        self.pc = pc
        self.sp = sp
        self.status = status

        # The rest of the registers
        self._registers = registers

    def __getattr__(self, name: str):
        if name in self._registers:
            return self._registers[name]

        raise AttributeError(f"No register named '{name}' in this core dump.")

    def __setattr__(self, name: str, value):
        if name.startswith("_") or name in ("arch", "pc", "sp", "status"):
            super().__setattr__(name, value)
        else:
            if hasattr(self, "_registers") and (name in self._registers):
                self._registers[name] = value
            else:
                super().__setattr__(name, value)

    def apply_to_cpu(self, cpu):
        """
        Set the CPU registers from this RegisterState object.
        """
        cpu.pc.set(self.pc)
        cpu.sp.set(self.sp)

        for reg_name, reg_val in self._registers.items():
            reg_name_lower = reg_name.lower()
            if hasattr(cpu, reg_name_lower):
                getattr(cpu, reg_name_lower).set(reg_val)



from ....c99.signal import Raise, Signal
from ..systemv import AArch64SysVModel


class AArch64SysVRaise(Raise, AArch64SysVModel):
    pass


class AArch64SysVSignal(Signal, AArch64SysVModel):
    pass


__all__ = [
    "AArch64SysVRaise",
    "AArch64SysVSignal",
]



from ..... import platforms
from .....exceptions import ConfigurationError
from ..structs import ElfRela
from .rela import ElfRelocator

R_X86_64_64 = 1  # Direct 64-bit
R_X86_64_GLOB_DAT = 6  # Create GOT entry
R_X86_64_JUMP_SLOT = 7  # Create PLT entry
R_X86_64_RELATIVE = 8  # Adjust by program base
R_X86_64_NUM = 43  # This and higher aren't valid


class AMD64ElfRelocator(ElfRelocator):
    arch = platforms.Architecture.X86_64
    byteorder = platforms.Byteorder.LITTLE

    def _compute_value(self, rela: ElfRela):
        if (
            rela.type == R_X86_64_GLOB_DAT
            or rela.type == R_X86_64_JUMP_SLOT
            or rela.type == R_X86_64_RELATIVE
            or rela.type == R_X86_64_64
        ):
            # Different semantics, all behave the same
            val = rela.symbol.value + rela.symbol.baseaddr + rela.addend
            return val.to_bytes(8, "little")
        elif rela.type >= 0 and rela.type < R_X86_64_NUM:
            raise ConfigurationError(
                f"Valid, but unsupported relocation for {rela.symbol.name}: {rela.type}"
            )
        else:
            raise ConfigurationError(
                f"Invalid relocation type for {rela.symbol.name}: {hex(rela.type)}"
            )
